@@ -1,7 +1,6 @@
 package org.bouncycastle.crypto.signers;
 
 import java.security.SecureRandom;
-import java.util.Hashtable;
 
 import org.bouncycastle.crypto.AsymmetricBlockCipher;
 import org.bouncycastle.crypto.CipherParameters;
@@ -13,7 +12,6 @@ import org.bouncycastle.crypto.params.ParametersWithRandom;
 import org.bouncycastle.crypto.params.ParametersWithSalt;
 import org.bouncycastle.crypto.params.RSAKeyParameters;
 import org.bouncycastle.util.Arrays;
-import org.bouncycastle.util.Integers;
 
 /**
  * ISO9796-2 - mechanism using a hash function with recovery (scheme 2 and 3).
@@ -24,30 +22,23 @@ import org.bouncycastle.util.Integers;
 public class ISO9796d2PSSSigner
     implements SignerWithRecovery
 {
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_IMPLICIT    = 0xBC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_RIPEMD160   = 0x31CC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_RIPEMD128   = 0x32CC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_SHA1        = 0x33CC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_SHA256      = 0x34CC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_SHA512      = 0x35CC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_SHA384      = 0x36CC;
+    /** @deprecated use ISOTrailers */
     static final public int   TRAILER_WHIRLPOOL   = 0x37CC;
 
-    private static Hashtable trailerMap          = new Hashtable();
-
-    static
-    {
-        trailerMap.put("RIPEMD128", Integers.valueOf(TRAILER_RIPEMD128));
-        trailerMap.put("RIPEMD160", Integers.valueOf(TRAILER_RIPEMD160));
-
-        trailerMap.put("SHA-1", Integers.valueOf(TRAILER_SHA1));
-        trailerMap.put("SHA-256", Integers.valueOf(TRAILER_SHA256));
-        trailerMap.put("SHA-384", Integers.valueOf(TRAILER_SHA384));
-        trailerMap.put("SHA-512", Integers.valueOf(TRAILER_SHA512));
-
-        trailerMap.put("Whirlpool", Integers.valueOf(TRAILER_WHIRLPOOL));
-    }
-
     private Digest digest;
     private AsymmetricBlockCipher cipher;
 
@@ -70,8 +61,7 @@ public class ISO9796d2PSSSigner
     private int preTLength;
 
     /**
-     * Generate a signer for the with either implicit or explicit trailers
-     * for ISO9796-2, scheme 2 or 3.
+     * Generate a signer with either implicit or explicit trailers for ISO9796-2, scheme 2 or 3.
      *
      * @param cipher     base cipher to use for signature creation/verification
      * @param digest     digest to use.
@@ -91,11 +81,11 @@ public class ISO9796d2PSSSigner
 
         if (implicit)
         {
-            trailer = TRAILER_IMPLICIT;
+            trailer = ISOTrailers.TRAILER_IMPLICIT;
         }
         else
         {
-            Integer trailerObj = (Integer)trailerMap.get(digest.getAlgorithmName());
+            Integer trailerObj = ISOTrailers.getTrailer(digest);
 
             if (trailerObj != null)
             {
@@ -103,7 +93,7 @@ public class ISO9796d2PSSSigner
             }
             else
             {
-                throw new IllegalArgumentException("no valid trailer for digest");
+                throw new IllegalArgumentException("no valid trailer for digest: " + digest.getAlgorithmName());
             }
         }
     }
@@ -178,7 +168,7 @@ public class ISO9796d2PSSSigner
 
         block = new byte[(keyBits + 7) / 8];
 
-        if (trailer == TRAILER_IMPLICIT)
+        if (trailer == ISOTrailers.TRAILER_IMPLICIT)
         {
             mBuf = new byte[block.length - digest.getDigestSize() - lengthOfSalt - 1 - 1];
         }
@@ -254,7 +244,7 @@ public class ISO9796d2PSSSigner
         {
             int sigTrail = ((block[block.length - 2] & 0xFF) << 8) | (block[block.length - 1] & 0xFF);
 
-            Integer trailerObj = (Integer)trailerMap.get(digest.getAlgorithmName());
+            Integer trailerObj = ISOTrailers.getTrailer(digest);
 
             if (trailerObj != null)
             {
@@ -426,7 +416,7 @@ public class ISO9796d2PSSSigner
         digest.doFinal(hash, 0);
 
         int tLength = 2;
-        if (trailer == TRAILER_IMPLICIT)
+        if (trailer == ISOTrailers.TRAILER_IMPLICIT)
         {
             tLength = 1;
         }
@@ -446,9 +436,9 @@ public class ISO9796d2PSSSigner
 
         System.arraycopy(hash, 0, block, block.length - hLen - tLength, hLen);
 
-        if (trailer == TRAILER_IMPLICIT)
+        if (trailer == ISOTrailers.TRAILER_IMPLICIT)
         {
-            block[block.length - 1] = (byte)TRAILER_IMPLICIT;
+            block[block.length - 1] = (byte)ISOTrailers.TRAILER_IMPLICIT;
         }
         else
         {
@@ -460,6 +450,11 @@ public class ISO9796d2PSSSigner
 
         byte[] b = cipher.processBlock(block, 0, block.length);
 
+        recoveredMessage = new byte[messageLength];
+
+        fullMessage = (messageLength <= mBuf.length);
+        System.arraycopy(mBuf, 0, recoveredMessage, 0, recoveredMessage.length);
+
         clearBlock(mBuf);
         clearBlock(block);
         messageLength = 0;
@@ -526,7 +521,14 @@ public class ISO9796d2PSSSigner
         digest.update(m2Hash, 0, m2Hash.length);
 
         // Update for the salt
-        digest.update(block, mStart + recoveredMessage.length, saltLength);
+        if (standardSalt != null)
+        {
+            digest.update(standardSalt, 0, standardSalt.length);
+        }
+        else
+        {
+            digest.update(block, mStart + recoveredMessage.length, saltLength);
+        }
 
         byte[] hash = new byte[digest.getDigestSize()];
         digest.doFinal(hash, 0);
@@ -582,8 +584,10 @@ public class ISO9796d2PSSSigner
         return fullMessage;
     }
 
+
     /**
-     * Return a reference to the recoveredMessage message.
+     * Return a reference to the recoveredMessage message, either as it was added
+     * to a just generated signature, or extracted from a verified one.
      *
      * @return the full/partial recoveredMessage message.
      * @see org.bouncycastle.crypto.SignerWithRecovery#getRecoveredMessage()
