@@ -45,7 +45,7 @@ public class DTLSClientProtocol
         DTLSRecordLayer recordLayer = new DTLSRecordLayer(transport, state.clientContext, client, ContentType.handshake);
 
         TlsSession sessionToResume = state.client.getSessionToResume();
-        if (sessionToResume != null)
+        if (sessionToResume != null && sessionToResume.isResumable())
         {
             SessionParameters sessionParameters = sessionToResume.exportSessionParameters();
             if (sessionParameters != null)
@@ -61,19 +61,29 @@ public class DTLSClientProtocol
         }
         catch (TlsFatalAlert fatalAlert)
         {
-            recordLayer.fail(fatalAlert.getAlertDescription());
+            abortClientHandshake(state, recordLayer, fatalAlert.getAlertDescription());
             throw fatalAlert;
         }
         catch (IOException e)
         {
-            recordLayer.fail(AlertDescription.internal_error);
+            abortClientHandshake(state, recordLayer, AlertDescription.internal_error);
             throw e;
         }
         catch (RuntimeException e)
         {
-            recordLayer.fail(AlertDescription.internal_error);
-            throw new TlsFatalAlert(AlertDescription.internal_error);
+            abortClientHandshake(state, recordLayer, AlertDescription.internal_error);
+            throw new TlsFatalAlert(AlertDescription.internal_error, e);
         }
+        finally
+        {
+            securityParameters.clear();
+        }
+    }
+
+    protected void abortClientHandshake(ClientHandshakeState state, DTLSRecordLayer recordLayer, short alertDescription)
+    {
+        recordLayer.fail(alertDescription);
+        invalidateSession(state);
     }
 
     protected DTLSTransport clientHandshake(ClientHandshakeState state, DTLSRecordLayer recordLayer)
@@ -83,13 +93,16 @@ public class DTLSClientProtocol
         DTLSReliableHandshake handshake = new DTLSReliableHandshake(state.clientContext, recordLayer);
 
         byte[] clientHelloBody = generateClientHello(state, state.client);
+
+        recordLayer.setWriteVersion(ProtocolVersion.DTLSv10);
+
         handshake.sendMessage(HandshakeType.client_hello, clientHelloBody);
 
         DTLSReliableHandshake.Message serverMessage = handshake.receiveMessage();
 
         while (serverMessage.getType() == HandshakeType.hello_verify_request)
         {
-            ProtocolVersion recordLayerVersion = recordLayer.resetDiscoveredPeerVersion();
+            ProtocolVersion recordLayerVersion = recordLayer.getReadVersion();
             ProtocolVersion client_version = state.clientContext.getClientVersion();
 
             /*
@@ -103,6 +116,8 @@ public class DTLSClientProtocol
                 throw new TlsFatalAlert(AlertDescription.illegal_parameter);
             }
 
+            recordLayer.setReadVersion(null);
+
             byte[] cookie = processHelloVerifyRequest(state, serverMessage.getBody());
             byte[] patched = patchClientHelloWithCookie(clientHelloBody, cookie);
 
@@ -114,7 +129,9 @@ public class DTLSClientProtocol
 
         if (serverMessage.getType() == HandshakeType.server_hello)
         {
-            reportServerVersion(state, recordLayer.getDiscoveredPeerVersion());
+            ProtocolVersion recordLayerVersion = recordLayer.getReadVersion();
+            reportServerVersion(state, recordLayerVersion);
+            recordLayer.setWriteVersion(recordLayerVersion);
 
             processServerHello(state, serverMessage.getBody());
         }
@@ -123,35 +140,12 @@ public class DTLSClientProtocol
             throw new TlsFatalAlert(AlertDescription.unexpected_message);
         }
 
-        if (state.maxFragmentLength >= 0)
-        {
-            int plainTextLimit = 1 << (8 + state.maxFragmentLength);
-            recordLayer.setPlaintextLimit(plainTextLimit);
-        }
-
-        securityParameters.cipherSuite = state.selectedCipherSuite;
-        securityParameters.compressionAlgorithm = state.selectedCompressionMethod;
-        securityParameters.prfAlgorithm = TlsProtocol.getPRFAlgorithm(state.clientContext, state.selectedCipherSuite);
-
-        /*
-         * RFC 5264 7.4.9. Any cipher suite which does not explicitly specify verify_data_length has
-         * a verify_data_length equal to 12. This includes all existing cipher suites.
-         */
-        securityParameters.verifyDataLength = 12;
-
         handshake.notifyHelloComplete();
 
-        boolean resumedSession = state.selectedSessionID.length > 0 && state.tlsSession != null
-            && Arrays.areEqual(state.selectedSessionID, state.tlsSession.getSessionID());
+        applyMaxFragmentLengthExtension(recordLayer, securityParameters.maxFragmentLength);
 
-        if (resumedSession)
+        if (state.resumedSession)
         {
-            if (securityParameters.getCipherSuite() != state.sessionParameters.getCipherSuite()
-                || securityParameters.getCompressionAlgorithm() != state.sessionParameters.getCompressionAlgorithm())
-            {
-                throw new TlsFatalAlert(AlertDescription.illegal_parameter);
-            }
-
             securityParameters.masterSecret = Arrays.clone(state.sessionParameters.getMasterSecret());
             recordLayer.initPendingEpoch(state.client.getCipher());
 
@@ -311,11 +305,12 @@ public class DTLSClientProtocol
         byte[] clientKeyExchangeBody = generateClientKeyExchange(state);
         handshake.sendMessage(HandshakeType.client_key_exchange, clientKeyExchangeBody);
 
+        TlsHandshakeHash prepareFinishHash = handshake.prepareToFinish();
+        securityParameters.sessionHash = TlsProtocol.getCurrentPRFHash(state.clientContext, prepareFinishHash, null);
+
         TlsProtocol.establishMasterSecret(state.clientContext, state.keyExchange);
         recordLayer.initPendingEpoch(state.client.getCipher());
 
-        TlsHandshakeHash prepareFinishHash = handshake.prepareToFinish();
-
         if (state.clientCredentials != null && state.clientCredentials instanceof TlsSignerCredentials)
         {
             TlsSignerCredentials signerCredentials = (TlsSignerCredentials)state.clientCredentials;
@@ -323,23 +318,17 @@ public class DTLSClientProtocol
             /*
              * RFC 5246 4.7. digitally-signed element needs SignatureAndHashAlgorithm from TLS 1.2
              */
-            SignatureAndHashAlgorithm signatureAndHashAlgorithm;
+            SignatureAndHashAlgorithm signatureAndHashAlgorithm = TlsUtils.getSignatureAndHashAlgorithm(
+                state.clientContext, signerCredentials);
+
             byte[] hash;
-
-            if (TlsUtils.isTLSv12(state.clientContext))
+            if (signatureAndHashAlgorithm == null)
             {
-                signatureAndHashAlgorithm = signerCredentials.getSignatureAndHashAlgorithm();
-                if (signatureAndHashAlgorithm == null)
-                {
-                    throw new TlsFatalAlert(AlertDescription.internal_error);
-                }
-
-                hash = prepareFinishHash.getFinalHash(signatureAndHashAlgorithm.getHash());
+                hash = securityParameters.getSessionHash();
             }
             else
             {
-                signatureAndHashAlgorithm = null;
-                hash = TlsProtocol.getCurrentPRFHash(state.clientContext, prepareFinishHash, null);
+                hash = prepareFinishHash.getFinalHash(signatureAndHashAlgorithm.getHash());
             }
 
             byte[] signature = signerCredentials.generateCertificateSignature(hash);
@@ -376,10 +365,14 @@ public class DTLSClientProtocol
         if (state.tlsSession != null)
         {
             state.sessionParameters = new SessionParameters.Builder()
-                .setCipherSuite(securityParameters.cipherSuite)
-                .setCompressionAlgorithm(securityParameters.compressionAlgorithm)
-                .setMasterSecret(securityParameters.masterSecret)
+                .setCipherSuite(securityParameters.getCipherSuite())
+                .setCompressionAlgorithm(securityParameters.getCompressionAlgorithm())
+                .setMasterSecret(securityParameters.getMasterSecret())
                 .setPeerCertificate(serverCertificate)
+                .setPSKIdentity(securityParameters.getPSKIdentity())
+                .setSRPIdentity(securityParameters.getSRPIdentity())
+                // TODO Consider filtering extensions that aren't relevant to resumed sessions
+                .setServerExtensions(state.serverExtensions)
                 .build();
 
             state.tlsSession = TlsUtils.importSession(state.tlsSession.getSessionID(), state.sessionParameters);
@@ -411,10 +404,13 @@ public class DTLSClientProtocol
             throw new TlsFatalAlert(AlertDescription.internal_error);
         }
 
-        state.clientContext.setClientVersion(client_version);
+        TlsClientContextImpl context = state.clientContext;
+
+        context.setClientVersion(client_version);
         TlsUtils.writeVersion(client_version, buf);
 
-        buf.write(state.clientContext.getSecurityParameters().getClientRandom());
+        SecurityParameters securityParameters = context.getSecurityParameters();
+        buf.write(securityParameters.getClientRandom());
 
         // Session ID
         byte[] session_id = TlsUtils.EMPTY_BYTES;
@@ -431,6 +427,8 @@ public class DTLSClientProtocol
         // Cookie
         TlsUtils.writeOpaque8(TlsUtils.EMPTY_BYTES, buf);
 
+        boolean fallback = client.isFallback();
+
         /*
          * Cipher suites
          */
@@ -449,14 +447,26 @@ public class DTLSClientProtocol
             byte[] renegExtData = TlsUtils.getExtensionData(state.clientExtensions, TlsProtocol.EXT_RenegotiationInfo);
             boolean noRenegExt = (null == renegExtData);
 
-            boolean noSCSV = !Arrays.contains(state.offeredCipherSuites, CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV);
+            boolean noRenegSCSV = !Arrays.contains(state.offeredCipherSuites, CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV);
 
-            if (noRenegExt && noSCSV)
+            if (noRenegExt && noRenegSCSV)
             {
                 // TODO Consider whether to default to a client extension instead
                 state.offeredCipherSuites = Arrays.append(state.offeredCipherSuites, CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV);
             }
 
+            /*
+             * RFC 7507 4. If a client sends a ClientHello.client_version containing a lower value
+             * than the latest (highest-valued) version supported by the client, it SHOULD include
+             * the TLS_FALLBACK_SCSV cipher suite value in ClientHello.cipher_suites [..]. (The
+             * client SHOULD put TLS_FALLBACK_SCSV after all cipher suites that it actually intends
+             * to negotiate.)
+             */
+            if (fallback && !Arrays.contains(state.offeredCipherSuites, CipherSuite.TLS_FALLBACK_SCSV))
+            {
+                state.offeredCipherSuites = Arrays.append(state.offeredCipherSuites, CipherSuite.TLS_FALLBACK_SCSV);
+            }
+
             TlsUtils.writeUint16ArrayWithUint16Length(state.offeredCipherSuites, buf);
         }
 
@@ -606,8 +616,10 @@ public class DTLSClientProtocol
 
         ByteArrayInputStream buf = new ByteArrayInputStream(body);
 
-        ProtocolVersion server_version = TlsUtils.readVersion(buf);
-        reportServerVersion(state, server_version);
+        {
+            ProtocolVersion server_version = TlsUtils.readVersion(buf);
+            reportServerVersion(state, server_version);
+        }
 
         securityParameters.serverRandom = TlsUtils.readFully(32, buf);
 
@@ -617,26 +629,26 @@ public class DTLSClientProtocol
             throw new TlsFatalAlert(AlertDescription.illegal_parameter);
         }
         state.client.notifySessionID(state.selectedSessionID);
+        state.resumedSession = state.selectedSessionID.length > 0 && state.tlsSession != null
+            && Arrays.areEqual(state.selectedSessionID, state.tlsSession.getSessionID());
 
-        state.selectedCipherSuite = TlsUtils.readUint16(buf);
-        if (!Arrays.contains(state.offeredCipherSuites, state.selectedCipherSuite)
-            || state.selectedCipherSuite == CipherSuite.TLS_NULL_WITH_NULL_NULL
-            || state.selectedCipherSuite == CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV
-            || !TlsUtils.isValidCipherSuiteForVersion(state.selectedCipherSuite, server_version))
+        int selectedCipherSuite = TlsUtils.readUint16(buf);
+        if (!Arrays.contains(state.offeredCipherSuites, selectedCipherSuite)
+            || selectedCipherSuite == CipherSuite.TLS_NULL_WITH_NULL_NULL
+            || CipherSuite.isSCSV(selectedCipherSuite)
+            || !TlsUtils.isValidCipherSuiteForVersion(selectedCipherSuite, state.clientContext.getServerVersion()))
         {
             throw new TlsFatalAlert(AlertDescription.illegal_parameter);
         }
+        validateSelectedCipherSuite(selectedCipherSuite, AlertDescription.illegal_parameter);
+        state.client.notifySelectedCipherSuite(selectedCipherSuite);
 
-        validateSelectedCipherSuite(state.selectedCipherSuite, AlertDescription.illegal_parameter);
-
-        state.client.notifySelectedCipherSuite(state.selectedCipherSuite);
-
-        state.selectedCompressionMethod = TlsUtils.readUint8(buf);
-        if (!Arrays.contains(state.offeredCompressionMethods, state.selectedCompressionMethod))
+        short selectedCompressionMethod = TlsUtils.readUint8(buf);
+        if (!Arrays.contains(state.offeredCompressionMethods, selectedCompressionMethod))
         {
             throw new TlsFatalAlert(AlertDescription.illegal_parameter);
         }
-        state.client.notifySelectedCompressionMethod(state.selectedCompressionMethod);
+        state.client.notifySelectedCompressionMethod(selectedCompressionMethod);
 
         /*
          * RFC3546 2.2 The extended server hello message format MAY be sent in place of the server
@@ -654,101 +666,162 @@ public class DTLSClientProtocol
          */
 
         // Integer -> byte[]
-        Hashtable serverExtensions = TlsProtocol.readExtensions(buf);
+        state.serverExtensions = TlsProtocol.readExtensions(buf);
 
         /*
          * RFC 3546 2.2 Note that the extended server hello message is only sent in response to an
          * extended client hello message. However, see RFC 5746 exception below. We always include
          * the SCSV, so an Extended Server Hello is always allowed.
          */
-        if (serverExtensions != null)
+        if (state.serverExtensions != null)
         {
-            Enumeration e = serverExtensions.keys();
+            Enumeration e = state.serverExtensions.keys();
             while (e.hasMoreElements())
             {
                 Integer extType = (Integer)e.nextElement();
 
                 /*
-                 * RFC 5746 Note that sending a "renegotiation_info" extension in response to a
+                 * RFC 5746 3.6. Note that sending a "renegotiation_info" extension in response to a
                  * ClientHello containing only the SCSV is an explicit exception to the prohibition
                  * in RFC 5246, Section 7.4.1.4, on the server sending unsolicited extensions and is
                  * only allowed because the client is signaling its willingness to receive the
-                 * extension via the TLS_EMPTY_RENEGOTIATION_INFO_SCSV SCSV. TLS implementations
-                 * MUST continue to comply with Section 7.4.1.4 for all other extensions.
+                 * extension via the TLS_EMPTY_RENEGOTIATION_INFO_SCSV SCSV.
                  */
-                if (!extType.equals(TlsProtocol.EXT_RenegotiationInfo)
-                    && null == TlsUtils.getExtensionData(state.clientExtensions, extType))
+                if (extType.equals(TlsProtocol.EXT_RenegotiationInfo))
+                {
+                    continue;
+                }
+
+                /*
+                 * RFC 5246 7.4.1.4 An extension type MUST NOT appear in the ServerHello unless the
+                 * same extension type appeared in the corresponding ClientHello. If a client
+                 * receives an extension type in ServerHello that it did not request in the
+                 * associated ClientHello, it MUST abort the handshake with an unsupported_extension
+                 * fatal alert.
+                 */
+                if (null == TlsUtils.getExtensionData(state.clientExtensions, extType))
                 {
-                    /*
-                     * RFC 3546 2.3 Note that for all extension types (including those defined in
-                     * future), the extension type MUST NOT appear in the extended server hello
-                     * unless the same extension type appeared in the corresponding client hello.
-                     * Thus clients MUST abort the handshake if they receive an extension type in
-                     * the extended server hello that they did not request in the associated
-                     * (extended) client hello.
-                     */
                     throw new TlsFatalAlert(AlertDescription.unsupported_extension);
                 }
-            }
 
-            /*
-             * RFC 5746 3.4. Client Behavior: Initial Handshake
-             */
-            {
                 /*
-                 * When a ServerHello is received, the client MUST check if it includes the
-                 * "renegotiation_info" extension:
+                 * RFC 3546 2.3. If [...] the older session is resumed, then the server MUST ignore
+                 * extensions appearing in the client hello, and send a server hello containing no
+                 * extensions[.]
                  */
-                byte[] renegExtData = (byte[])serverExtensions.get(TlsProtocol.EXT_RenegotiationInfo);
-                if (renegExtData != null)
+                if (state.resumedSession)
                 {
-                    /*
-                     * If the extension is present, set the secure_renegotiation flag to TRUE. The
-                     * client MUST then verify that the length of the "renegotiated_connection"
-                     * field is zero, and if it is not, MUST abort the handshake (by sending a fatal
-                     * handshake_failure alert).
-                     */
-                    state.secure_renegotiation = true;
-
-                    if (!Arrays.constantTimeAreEqual(renegExtData,
-                        TlsProtocol.createRenegotiationInfo(TlsUtils.EMPTY_BYTES)))
-                    {
-                        throw new TlsFatalAlert(AlertDescription.handshake_failure);
-                    }
+                    // TODO[compat-gnutls] GnuTLS test server sends server extensions e.g. ec_point_formats
+                    // TODO[compat-openssl] OpenSSL test server sends server extensions e.g. ec_point_formats
+                    // TODO[compat-polarssl] PolarSSL test server sends server extensions e.g. ec_point_formats
+//                    throw new TlsFatalAlert(AlertDescription.illegal_parameter);
                 }
             }
+        }
 
+        /*
+         * RFC 5746 3.4. Client Behavior: Initial Handshake
+         */
+        {
             /*
-             * draft-ietf-tls-encrypt-then-mac-03 3. If a server receives an encrypt-then-MAC
-             * request extension from a client and then selects a stream or AEAD cipher suite, it
-             * MUST NOT send an encrypt-then-MAC response extension back to the client.
+             * When a ServerHello is received, the client MUST check if it includes the
+             * "renegotiation_info" extension:
              */
-            boolean serverSentEncryptThenMAC = TlsExtensionsUtils.hasEncryptThenMACExtension(serverExtensions);
-            if (serverSentEncryptThenMAC && !TlsUtils.isBlockCipherSuite(state.selectedCipherSuite))
+            byte[] renegExtData = TlsUtils.getExtensionData(state.serverExtensions, TlsProtocol.EXT_RenegotiationInfo);
+            if (renegExtData != null)
+            {
+                /*
+                 * If the extension is present, set the secure_renegotiation flag to TRUE. The
+                 * client MUST then verify that the length of the "renegotiated_connection"
+                 * field is zero, and if it is not, MUST abort the handshake (by sending a fatal
+                 * handshake_failure alert).
+                 */
+                state.secure_renegotiation = true;
+
+                if (!Arrays.constantTimeAreEqual(renegExtData,
+                    TlsProtocol.createRenegotiationInfo(TlsUtils.EMPTY_BYTES)))
+                {
+                    throw new TlsFatalAlert(AlertDescription.handshake_failure);
+                }
+            }
+        }
+
+        // TODO[compat-gnutls] GnuTLS test server fails to send renegotiation_info extension when resuming
+        state.client.notifySecureRenegotiation(state.secure_renegotiation);
+
+        Hashtable sessionClientExtensions = state.clientExtensions, sessionServerExtensions = state.serverExtensions;
+        if (state.resumedSession)
+        {
+            if (selectedCipherSuite != state.sessionParameters.getCipherSuite()
+                || selectedCompressionMethod != state.sessionParameters.getCompressionAlgorithm())
             {
                 throw new TlsFatalAlert(AlertDescription.illegal_parameter);
             }
 
-            securityParameters.encryptThenMAC = serverSentEncryptThenMAC;
-
-            state.maxFragmentLength = evaluateMaxFragmentLengthExtension(state.clientExtensions, serverExtensions,
-                AlertDescription.illegal_parameter);
-
-            securityParameters.truncatedHMac = TlsExtensionsUtils.hasTruncatedHMacExtension(serverExtensions);
-
-            state.allowCertificateStatus = TlsUtils.hasExpectedEmptyExtensionData(serverExtensions,
-                TlsExtensionsUtils.EXT_status_request, AlertDescription.illegal_parameter);
-
-            state.expectSessionTicket = TlsUtils.hasExpectedEmptyExtensionData(serverExtensions,
-                TlsProtocol.EXT_SessionTicket, AlertDescription.illegal_parameter);
+            sessionClientExtensions = null;
+            sessionServerExtensions = state.sessionParameters.readServerExtensions();
         }
 
-        state.client.notifySecureRenegotiation(state.secure_renegotiation);
+        securityParameters.cipherSuite = selectedCipherSuite;
+        securityParameters.compressionAlgorithm = selectedCompressionMethod;
 
-        if (state.clientExtensions != null)
+        if (sessionServerExtensions != null)
         {
-            state.client.processServerExtensions(serverExtensions);
+            {
+                /*
+                 * RFC 7366 3. If a server receives an encrypt-then-MAC request extension from a client
+                 * and then selects a stream or Authenticated Encryption with Associated Data (AEAD)
+                 * ciphersuite, it MUST NOT send an encrypt-then-MAC response extension back to the
+                 * client.
+                 */
+                boolean serverSentEncryptThenMAC = TlsExtensionsUtils.hasEncryptThenMACExtension(sessionServerExtensions);
+                if (serverSentEncryptThenMAC && !TlsUtils.isBlockCipherSuite(securityParameters.getCipherSuite()))
+                {
+                    throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+                }
+                securityParameters.encryptThenMAC = serverSentEncryptThenMAC;
+            }
+
+            securityParameters.extendedMasterSecret = TlsExtensionsUtils.hasExtendedMasterSecretExtension(sessionServerExtensions);
+
+            securityParameters.maxFragmentLength = evaluateMaxFragmentLengthExtension(state.resumedSession,
+                sessionClientExtensions, sessionServerExtensions, AlertDescription.illegal_parameter);
+
+            securityParameters.truncatedHMac = TlsExtensionsUtils.hasTruncatedHMacExtension(sessionServerExtensions);
+
+            /*
+             * TODO It's surprising that there's no provision to allow a 'fresh' CertificateStatus to be
+             * sent in a session resumption handshake.
+             */
+            state.allowCertificateStatus = !state.resumedSession
+                && TlsUtils.hasExpectedEmptyExtensionData(sessionServerExtensions, TlsExtensionsUtils.EXT_status_request,
+                    AlertDescription.illegal_parameter);
+
+            state.expectSessionTicket = !state.resumedSession
+                && TlsUtils.hasExpectedEmptyExtensionData(sessionServerExtensions, TlsProtocol.EXT_SessionTicket,
+                    AlertDescription.illegal_parameter);
         }
+
+        /*
+         * TODO[session-hash]
+         * 
+         * draft-ietf-tls-session-hash-04 4. Clients and servers SHOULD NOT accept handshakes
+         * that do not use the extended master secret [..]. (and see 5.2, 5.3)
+         */
+
+        if (sessionClientExtensions != null)
+        {
+            state.client.processServerExtensions(sessionServerExtensions);
+        }
+
+        securityParameters.prfAlgorithm = TlsProtocol.getPRFAlgorithm(state.clientContext,
+            securityParameters.getCipherSuite());
+
+        /*
+         * RFC 5264 7.4.9. Any cipher suite which does not explicitly specify verify_data_length has
+         * a verify_data_length equal to 12. This includes all existing cipher suites.
+         */
+        securityParameters.verifyDataLength = 12;
     }
 
     protected void processServerKeyExchange(ClientHandshakeState state, byte[] body)
@@ -815,11 +888,10 @@ public class DTLSClientProtocol
         int[] offeredCipherSuites = null;
         short[] offeredCompressionMethods = null;
         Hashtable clientExtensions = null;
+        Hashtable serverExtensions = null;
         byte[] selectedSessionID = null;
-        int selectedCipherSuite = -1;
-        short selectedCompressionMethod = -1;
+        boolean resumedSession = false;
         boolean secure_renegotiation = false;
-        short maxFragmentLength = -1;
         boolean allowCertificateStatus = false;
         boolean expectSessionTicket = false;
         TlsKeyExchange keyExchange = null;
