@@ -9,14 +9,13 @@ import java.util.Enumeration;
 import java.util.Hashtable;
 import java.util.Vector;
 
-import org.bouncycastle.crypto.prng.ThreadedSeedGenerator;
 import org.bouncycastle.util.Arrays;
 
 public class TlsClientProtocol
     extends TlsProtocol
 {
     protected TlsClient tlsClient = null;
-    protected TlsClientContextImpl tlsClientContext = null;
+    TlsClientContextImpl tlsClientContext = null;
 
     protected byte[] selectedSessionID = null;
 
@@ -26,41 +25,46 @@ public class TlsClientProtocol
     protected CertificateStatus certificateStatus = null;
     protected CertificateRequest certificateRequest = null;
 
-    private static SecureRandom createSecureRandom()
-    {
-        /*
-         * We use our threaded seed generator to generate a good random seed. If the user has a
-         * better random seed, he should use the constructor with a SecureRandom.
-         */
-        ThreadedSeedGenerator tsg = new ThreadedSeedGenerator();
-        SecureRandom random = new SecureRandom();
-
-        /*
-         * Hopefully, 20 bytes in fast mode are good enough.
-         */
-        random.setSeed(tsg.generateSeed(20, true));
-
-        return random;
-    }
-
     /**
-     * @deprecated use alternate constructor taking an explicit {@link SecureRandom}
+     * Constructor for blocking mode.
+     * @param input The stream of data from the server
+     * @param output The stream of data to the server
+     * @param secureRandom Random number generator for various cryptographic functions
      */
-    public TlsClientProtocol(InputStream input, OutputStream output)
-    {
-        this(input, output, createSecureRandom());
-    }
-
     public TlsClientProtocol(InputStream input, OutputStream output, SecureRandom secureRandom)
     {
         super(input, output, secureRandom);
     }
 
     /**
-     * Initiates a TLS handshake in the role of client
+     * Constructor for non-blocking mode.<br>
+     * <br>
+     * When data is received, use {@link #offerInput(java.nio.ByteBuffer)} to
+     * provide the received ciphertext, then use
+     * {@link #readInput(byte[], int, int)} to read the corresponding cleartext.<br>
+     * <br>
+     * Similarly, when data needs to be sent, use
+     * {@link #offerOutput(byte[], int, int)} to provide the cleartext, then use
+     * {@link #readOutput(byte[], int, int)} to get the corresponding
+     * ciphertext.
+     * 
+     * @param secureRandom
+     *            Random number generator for various cryptographic functions
+     */
+    public TlsClientProtocol(SecureRandom secureRandom)
+    {
+        super(secureRandom);
+    }
+
+    /**
+     * Initiates a TLS handshake in the role of client.<br>
+     * <br>
+     * In blocking mode, this will not return until the handshake is complete.
+     * In non-blocking mode, use {@link TlsPeer#notifyHandshakeComplete()} to
+     * receive a callback when the handshake is complete.
      *
      * @param tlsClient The {@link TlsClient} to use for the handshake.
-     * @throws IOException If handshake was not successful.
+     * @throws IOException If in blocking mode and handshake was not successful.
      */
     public void connect(TlsClient tlsClient) throws IOException
     {
@@ -87,7 +91,7 @@ public class TlsClientProtocol
         this.recordStream.init(tlsClientContext);
 
         TlsSession sessionToResume = tlsClient.getSessionToResume();
-        if (sessionToResume != null)
+        if (sessionToResume != null && sessionToResume.isResumable())
         {
             SessionParameters sessionParameters = sessionToResume.exportSessionParameters();
             if (sessionParameters != null)
@@ -100,7 +104,7 @@ public class TlsClientProtocol
         sendClientHelloMessage();
         this.connection_state = CS_CLIENT_HELLO;
 
-        completeHandshake();
+        blockForHandshake();
     }
 
     protected void cleanupHandshake()
@@ -114,11 +118,16 @@ public class TlsClientProtocol
         this.certificateRequest = null;
     }
 
-    protected AbstractTlsContext getContext()
+    protected TlsContext getContext()
     {
         return tlsClientContext;
     }
 
+    AbstractTlsContext getContextAdmin()
+    {
+        return tlsClientContext;
+    }
+    
     protected TlsPeer getPeer()
     {
         return tlsClient;
@@ -143,6 +152,7 @@ public class TlsClientProtocol
             this.connection_state = CS_CLIENT_FINISHED;
             this.connection_state = CS_END;
 
+            completeHandshake();
             return;
         }
 
@@ -237,6 +247,8 @@ public class TlsClientProtocol
                 processFinishedMessage(buf);
                 this.connection_state = CS_SERVER_FINISHED;
                 this.connection_state = CS_END;
+
+                completeHandshake();
                 break;
             }
             default:
@@ -253,24 +265,10 @@ public class TlsClientProtocol
                 receiveServerHelloMessage(buf);
                 this.connection_state = CS_SERVER_HELLO;
 
-                if (this.securityParameters.maxFragmentLength >= 0)
-                {
-                    int plainTextLimit = 1 << (8 + this.securityParameters.maxFragmentLength);
-                    recordStream.setPlaintextLimit(plainTextLimit);
-                }
-
-                this.securityParameters.prfAlgorithm = getPRFAlgorithm(getContext(),
-                    this.securityParameters.getCipherSuite());
-
-                /*
-                 * RFC 5264 7.4.9. Any cipher suite which does not explicitly specify
-                 * verify_data_length has a verify_data_length equal to 12. This includes all
-                 * existing cipher suites.
-                 */
-                this.securityParameters.verifyDataLength = 12;
-
                 this.recordStream.notifyHelloComplete();
 
+                applyMaxFragmentLengthExtension();
+
                 if (this.resumedSession)
                 {
                     this.securityParameters.masterSecret = Arrays.clone(this.sessionParameters.getMasterSecret());
@@ -388,11 +386,12 @@ public class TlsClientProtocol
                 sendClientKeyExchangeMessage();
                 this.connection_state = CS_CLIENT_KEY_EXCHANGE;
 
+                TlsHandshakeHash prepareFinishHash = recordStream.prepareToFinish();
+                this.securityParameters.sessionHash = getCurrentPRFHash(getContext(), prepareFinishHash, null);
+
                 establishMasterSecret(getContext(), keyExchange);
                 recordStream.setPendingConnectionState(getPeer().getCompression(), getPeer().getCipher());
 
-                TlsHandshakeHash prepareFinishHash = recordStream.prepareToFinish();
-
                 if (clientCreds != null && clientCreds instanceof TlsSignerCredentials)
                 {
                     TlsSignerCredentials signerCredentials = (TlsSignerCredentials)clientCreds;
@@ -400,23 +399,17 @@ public class TlsClientProtocol
                     /*
                      * RFC 5246 4.7. digitally-signed element needs SignatureAndHashAlgorithm from TLS 1.2
                      */
-                    SignatureAndHashAlgorithm signatureAndHashAlgorithm;
+                    SignatureAndHashAlgorithm signatureAndHashAlgorithm = TlsUtils.getSignatureAndHashAlgorithm(
+                        getContext(), signerCredentials);
+
                     byte[] hash;
-
-                    if (TlsUtils.isTLSv12(getContext()))
+                    if (signatureAndHashAlgorithm == null)
                     {
-                        signatureAndHashAlgorithm = signerCredentials.getSignatureAndHashAlgorithm();
-                        if (signatureAndHashAlgorithm == null)
-                        {
-                            throw new TlsFatalAlert(AlertDescription.internal_error);
-                        }
-
-                        hash = prepareFinishHash.getFinalHash(signatureAndHashAlgorithm.getHash());
+                        hash = securityParameters.getSessionHash();
                     }
                     else
                     {
-                        signatureAndHashAlgorithm = null;
-                        hash = getCurrentPRFHash(getContext(), prepareFinishHash, null);
+                        hash = prepareFinishHash.getFinalHash(signatureAndHashAlgorithm.getHash());
                     }
 
                     byte[] signature = signerCredentials.generateCertificateSignature(hash);
@@ -428,12 +421,13 @@ public class TlsClientProtocol
 
                 sendChangeCipherSpecMessage();
                 sendFinishedMessage();
-                this.connection_state = CS_CLIENT_FINISHED;
                 break;
             }
             default:
                 throw new TlsFatalAlert(AlertDescription.handshake_failure);
             }
+
+            this.connection_state = CS_CLIENT_FINISHED;
             break;
         }
         case HandshakeType.server_key_exchange:
@@ -535,12 +529,14 @@ public class TlsClientProtocol
                 invalidateSession();
 
                 receiveNewSessionTicketMessage(buf);
-                this.connection_state = CS_SERVER_SESSION_TICKET;
                 break;
             }
             default:
                 throw new TlsFatalAlert(AlertDescription.unexpected_message);
             }
+
+            this.connection_state = CS_SERVER_SESSION_TICKET;
+            break;
         }
         case HandshakeType.hello_request:
         {
@@ -554,17 +550,7 @@ public class TlsClientProtocol
              */
             if (this.connection_state == CS_END)
             {
-                /*
-                 * RFC 5746 4.5 SSLv3 clients that refuse renegotiation SHOULD use a fatal
-                 * handshake_failure alert.
-                 */
-                if (TlsUtils.isSSL(getContext()))
-                {
-                    throw new TlsFatalAlert(AlertDescription.handshake_failure);
-                }
-
-                String message = "Renegotiation not supported";
-                raiseWarning(AlertDescription.no_renegotiation, message);
+                refuseRenegotiation();
             }
             break;
         }
@@ -592,7 +578,7 @@ public class TlsClientProtocol
     {
         NewSessionTicket newSessionTicket = NewSessionTicket.parse(buf);
 
-        TlsProtocol.assertEmpty(buf);
+        assertEmpty(buf);
 
         tlsClient.notifyNewSessionTicket(newSessionTicket);
     }
@@ -600,28 +586,30 @@ public class TlsClientProtocol
     protected void receiveServerHelloMessage(ByteArrayInputStream buf)
         throws IOException
     {
-        ProtocolVersion server_version = TlsUtils.readVersion(buf);
-        if (server_version.isDTLS())
         {
-            throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+            ProtocolVersion server_version = TlsUtils.readVersion(buf);
+            if (server_version.isDTLS())
+            {
+                throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+            }
+    
+            // Check that this matches what the server is sending in the record layer
+            if (!server_version.equals(this.recordStream.getReadVersion()))
+            {
+                throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+            }
+    
+            ProtocolVersion client_version = getContext().getClientVersion();
+            if (!server_version.isEqualOrEarlierVersionOf(client_version))
+            {
+                throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+            }
+    
+            this.recordStream.setWriteVersion(server_version);
+            getContextAdmin().setServerVersion(server_version);
+            this.tlsClient.notifyServerVersion(server_version);
         }
 
-        // Check that this matches what the server is sending in the record layer
-        if (!server_version.equals(this.recordStream.getReadVersion()))
-        {
-            throw new TlsFatalAlert(AlertDescription.illegal_parameter);
-        }
-
-        ProtocolVersion client_version = getContext().getClientVersion();
-        if (!server_version.isEqualOrEarlierVersionOf(client_version))
-        {
-            throw new TlsFatalAlert(AlertDescription.illegal_parameter);
-        }
-
-        this.recordStream.setWriteVersion(server_version);
-        getContext().setServerVersion(server_version);
-        this.tlsClient.notifyServerVersion(server_version);
-
         /*
          * Read the server random
          */
@@ -632,9 +620,7 @@ public class TlsClientProtocol
         {
             throw new TlsFatalAlert(AlertDescription.illegal_parameter);
         }
-
         this.tlsClient.notifySessionID(this.selectedSessionID);
-
         this.resumedSession = this.selectedSessionID.length > 0 && this.tlsSession != null
             && Arrays.areEqual(this.selectedSessionID, this.tlsSession.getSessionID());
 
@@ -645,12 +631,11 @@ public class TlsClientProtocol
         int selectedCipherSuite = TlsUtils.readUint16(buf);
         if (!Arrays.contains(this.offeredCipherSuites, selectedCipherSuite)
             || selectedCipherSuite == CipherSuite.TLS_NULL_WITH_NULL_NULL
-            || selectedCipherSuite == CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV
-            || !TlsUtils.isValidCipherSuiteForVersion(selectedCipherSuite, server_version))
+            || CipherSuite.isSCSV(selectedCipherSuite)
+            || !TlsUtils.isValidCipherSuiteForVersion(selectedCipherSuite, getContext().getServerVersion()))
         {
             throw new TlsFatalAlert(AlertDescription.illegal_parameter);
         }
-
         this.tlsClient.notifySelectedCipherSuite(selectedCipherSuite);
 
         /*
@@ -662,7 +647,6 @@ public class TlsClientProtocol
         {
             throw new TlsFatalAlert(AlertDescription.illegal_parameter);
         }
-
         this.tlsClient.notifySelectedCompressionMethod(selectedCompressionMethod);
 
         /*
@@ -701,6 +685,18 @@ public class TlsClientProtocol
                     continue;
                 }
 
+                /*
+                 * RFC 5246 7.4.1.4 An extension type MUST NOT appear in the ServerHello unless the
+                 * same extension type appeared in the corresponding ClientHello. If a client
+                 * receives an extension type in ServerHello that it did not request in the
+                 * associated ClientHello, it MUST abort the handshake with an unsupported_extension
+                 * fatal alert.
+                 */
+                if (null == TlsUtils.getExtensionData(this.clientExtensions, extType))
+                {
+                    throw new TlsFatalAlert(AlertDescription.unsupported_extension);
+                }
+
                 /*
                  * RFC 3546 2.3. If [...] the older session is resumed, then the server MUST ignore
                  * extensions appearing in the client hello, and send a server hello containing no
@@ -713,18 +709,6 @@ public class TlsClientProtocol
                     // TODO[compat-polarssl] PolarSSL test server sends server extensions e.g. ec_point_formats
 //                    throw new TlsFatalAlert(AlertDescription.illegal_parameter);
                 }
-
-                /*
-                 * RFC 5246 7.4.1.4 An extension type MUST NOT appear in the ServerHello unless the
-                 * same extension type appeared in the corresponding ClientHello. If a client
-                 * receives an extension type in ServerHello that it did not request in the
-                 * associated ClientHello, it MUST abort the handshake with an unsupported_extension
-                 * fatal alert.
-                 */
-                if (null == TlsUtils.getExtensionData(this.clientExtensions, extType))
-                {
-                    throw new TlsFatalAlert(AlertDescription.unsupported_extension);
-                }
             }
         }
 
@@ -775,18 +759,22 @@ public class TlsClientProtocol
 
         if (sessionServerExtensions != null)
         {
-            /*
-             * draft-ietf-tls-encrypt-then-mac-03 3. If a server receives an encrypt-then-MAC
-             * request extension from a client and then selects a stream or AEAD cipher suite, it
-             * MUST NOT send an encrypt-then-MAC response extension back to the client.
-             */
-            boolean serverSentEncryptThenMAC = TlsExtensionsUtils.hasEncryptThenMACExtension(sessionServerExtensions);
-            if (serverSentEncryptThenMAC && !TlsUtils.isBlockCipherSuite(selectedCipherSuite))
             {
-                throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+                /*
+                 * RFC 7366 3. If a server receives an encrypt-then-MAC request extension from a client
+                 * and then selects a stream or Authenticated Encryption with Associated Data (AEAD)
+                 * ciphersuite, it MUST NOT send an encrypt-then-MAC response extension back to the
+                 * client.
+                 */
+                boolean serverSentEncryptThenMAC = TlsExtensionsUtils.hasEncryptThenMACExtension(sessionServerExtensions);
+                if (serverSentEncryptThenMAC && !TlsUtils.isBlockCipherSuite(selectedCipherSuite))
+                {
+                    throw new TlsFatalAlert(AlertDescription.illegal_parameter);
+                }
+                this.securityParameters.encryptThenMAC = serverSentEncryptThenMAC;
             }
 
-            this.securityParameters.encryptThenMAC = serverSentEncryptThenMAC;
+            this.securityParameters.extendedMasterSecret = TlsExtensionsUtils.hasExtendedMasterSecretExtension(sessionServerExtensions);
 
             this.securityParameters.maxFragmentLength = processMaxFragmentLengthExtension(sessionClientExtensions,
                 sessionServerExtensions, AlertDescription.illegal_parameter);
@@ -806,10 +794,27 @@ public class TlsClientProtocol
                     AlertDescription.illegal_parameter);
         }
 
+        /*
+         * TODO[session-hash]
+         * 
+         * draft-ietf-tls-session-hash-04 4. Clients and servers SHOULD NOT accept handshakes
+         * that do not use the extended master secret [..]. (and see 5.2, 5.3)
+         */
+        
         if (sessionClientExtensions != null)
         {
             this.tlsClient.processServerExtensions(sessionServerExtensions);
         }
+
+        this.securityParameters.prfAlgorithm = getPRFAlgorithm(getContext(),
+            this.securityParameters.getCipherSuite());
+
+        /*
+         * RFC 5264 7.4.9. Any cipher suite which does not explicitly specify
+         * verify_data_length has a verify_data_length equal to 12. This includes all
+         * existing cipher suites.
+         */
+        this.securityParameters.verifyDataLength = 12;
     }
 
     protected void sendCertificateVerifyMessage(DigitallySigned certificateVerify)
@@ -833,7 +838,7 @@ public class TlsClientProtocol
             throw new TlsFatalAlert(AlertDescription.internal_error);
         }
 
-        getContext().setClientVersion(client_version);
+        getContextAdmin().setClientVersion(client_version);
 
         /*
          * TODO RFC 5077 3.4. When presenting a ticket, the client MAY generate and include a
@@ -849,6 +854,8 @@ public class TlsClientProtocol
             }
         }
 
+        boolean fallback = this.tlsClient.isFallback();
+
         this.offeredCipherSuites = this.tlsClient.getCipherSuites();
 
         this.offeredCompressionMethods = this.tlsClient.getCompressionMethods();
@@ -882,9 +889,9 @@ public class TlsClientProtocol
             byte[] renegExtData = TlsUtils.getExtensionData(clientExtensions, EXT_RenegotiationInfo);
             boolean noRenegExt = (null == renegExtData);
 
-            boolean noSCSV = !Arrays.contains(offeredCipherSuites, CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV);
+            boolean noRenegSCSV = !Arrays.contains(offeredCipherSuites, CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV);
 
-            if (noRenegExt && noSCSV)
+            if (noRenegExt && noRenegSCSV)
             {
                 // TODO Consider whether to default to a client extension instead
 //                this.clientExtensions = TlsExtensionsUtils.ensureExtensionsInitialised(this.clientExtensions);
@@ -892,6 +899,18 @@ public class TlsClientProtocol
                 this.offeredCipherSuites = Arrays.append(offeredCipherSuites, CipherSuite.TLS_EMPTY_RENEGOTIATION_INFO_SCSV);
             }
 
+            /*
+             * RFC 7507 4. If a client sends a ClientHello.client_version containing a lower value
+             * than the latest (highest-valued) version supported by the client, it SHOULD include
+             * the TLS_FALLBACK_SCSV cipher suite value in ClientHello.cipher_suites [..]. (The
+             * client SHOULD put TLS_FALLBACK_SCSV after all cipher suites that it actually intends
+             * to negotiate.)
+             */
+            if (fallback && !Arrays.contains(offeredCipherSuites, CipherSuite.TLS_FALLBACK_SCSV))
+            {
+                this.offeredCipherSuites = Arrays.append(offeredCipherSuites, CipherSuite.TLS_FALLBACK_SCSV);
+            }
+
             TlsUtils.writeUint16ArrayWithUint16Length(offeredCipherSuites, message);
         }
 
