@@ -16,9 +16,6 @@ import org.bouncycastle.crypto.prng.RandomGenerator;
 import org.bouncycastle.util.Arrays;
 import org.bouncycastle.util.Integers;
 
-/**
- * An implementation of all high level protocols in TLS 1.0/1.1.
- */
 public abstract class TlsProtocol
 {
     protected static final Integer EXT_RenegotiationInfo = Integers.valueOf(ExtensionType.renegotiation_info);
@@ -47,6 +44,13 @@ public abstract class TlsProtocol
     protected static final short CS_SERVER_FINISHED = 15;
     protected static final short CS_END = 16;
 
+    /*
+     * Different modes to handle the known IV weakness
+     */
+    protected static final short ADS_MODE_1_Nsub1 = 0; // 1/n-1 record splitting
+    protected static final short ADS_MODE_0_N = 1; // 0/n record splitting
+    protected static final short ADS_MODE_0_N_FIRSTONLY = 2; // 0/n record splitting on first data fragment only
+
     /*
      * Queues for data from some protocols.
      */
@@ -58,7 +62,7 @@ public abstract class TlsProtocol
     /*
      * The Record Stream we use
      */
-    protected RecordStream recordStream;
+    RecordStream recordStream;
     protected SecureRandom secureRandom;
 
     private TlsInputStream tlsInputStream = null;
@@ -67,7 +71,8 @@ public abstract class TlsProtocol
     private volatile boolean closed = false;
     private volatile boolean failedWithError = false;
     private volatile boolean appDataReady = false;
-    private volatile boolean splitApplicationDataRecords = true;
+    private volatile boolean appDataSplitEnabled = true;
+    private volatile int appDataSplitMode = ADS_MODE_1_Nsub1;
     private byte[] expected_verify_data = null;
 
     protected TlsSession tlsSession = null;
@@ -87,13 +92,29 @@ public abstract class TlsProtocol
     protected boolean allowCertificateStatus = false;
     protected boolean expectSessionTicket = false;
 
+    protected boolean blocking;
+    protected ByteQueueInputStream inputBuffers;
+    protected ByteQueueOutputStream outputBuffer;
+    
     public TlsProtocol(InputStream input, OutputStream output, SecureRandom secureRandom)
     {
+        this.blocking = true;
         this.recordStream = new RecordStream(this, input, output);
         this.secureRandom = secureRandom;
     }
+    
+    public TlsProtocol(SecureRandom secureRandom)
+    {
+        this.blocking = false;
+        this.inputBuffers = new ByteQueueInputStream();
+        this.outputBuffer = new ByteQueueOutputStream();
+        this.recordStream = new RecordStream(this, inputBuffers, outputBuffer);
+        this.secureRandom = secureRandom;
+    }
 
-    protected abstract AbstractTlsContext getContext();
+    protected abstract TlsContext getContext();
+
+    abstract AbstractTlsContext getContextAdmin();
 
     protected abstract TlsPeer getPeer();
 
@@ -109,6 +130,30 @@ public abstract class TlsProtocol
     {
     }
 
+    protected void applyMaxFragmentLengthExtension()
+        throws IOException
+    {
+        if (securityParameters.maxFragmentLength >= 0)
+        {
+            if (!MaxFragmentLength.isValid(securityParameters.maxFragmentLength))
+            {
+                throw new TlsFatalAlert(AlertDescription.internal_error); 
+            }
+    
+            int plainTextLimit = 1 << (8 + securityParameters.maxFragmentLength);
+            recordStream.setPlaintextLimit(plainTextLimit);
+        }
+    }
+
+    protected void checkReceivedChangeCipherSpec(boolean expected)
+        throws IOException
+    {
+        if (expected != receivedChangeCipherSpec)
+        {
+            throw new TlsFatalAlert(AlertDescription.unexpected_message);
+        }
+    }
+
     protected void cleanupHandshake()
     {
         if (this.expected_verify_data != null)
@@ -131,15 +176,11 @@ public abstract class TlsProtocol
         this.allowCertificateStatus = false;
         this.expectSessionTicket = false;
     }
-
-    protected void completeHandshake()
-        throws IOException
+    
+    protected void blockForHandshake() throws IOException
     {
-        try
+        if (blocking)
         {
-            /*
-             * We will now read data, until we have completed the handshake.
-             */
             while (this.connection_state != CS_END)
             {
                 if (this.closed)
@@ -149,10 +190,17 @@ public abstract class TlsProtocol
 
                 safeReadRecord();
             }
+        }
+    }
 
+    protected void completeHandshake()
+        throws IOException
+    {
+        try
+        {
             this.recordStream.finaliseHandshake();
 
-            this.splitApplicationDataRecords = !TlsUtils.isTLSv11(getContext());
+            this.appDataSplitEnabled = !TlsUtils.isTLSv11(getContext());
 
             /*
              * If this was an initial handshake, we are now ready to send and receive application data.
@@ -161,8 +209,11 @@ public abstract class TlsProtocol
             {
                 this.appDataReady = true;
 
-                this.tlsInputStream = new TlsInputStream(this);
-                this.tlsOutputStream = new TlsOutputStream(this);
+                if (blocking)
+                {
+                    this.tlsInputStream = new TlsInputStream(this);
+                    this.tlsOutputStream = new TlsOutputStream(this);
+                }
             }
 
             if (this.tlsSession != null)
@@ -170,10 +221,12 @@ public abstract class TlsProtocol
                 if (this.sessionParameters == null)
                 {
                     this.sessionParameters = new SessionParameters.Builder()
-                        .setCipherSuite(this.securityParameters.cipherSuite)
-                        .setCompressionAlgorithm(this.securityParameters.compressionAlgorithm)
-                        .setMasterSecret(this.securityParameters.masterSecret)
+                        .setCipherSuite(this.securityParameters.getCipherSuite())
+                        .setCompressionAlgorithm(this.securityParameters.getCompressionAlgorithm())
+                        .setMasterSecret(this.securityParameters.getMasterSecret())
                         .setPeerCertificate(this.peerCertificate)
+                        .setPSKIdentity(this.securityParameters.getPSKIdentity())
+                        .setSRPIdentity(this.securityParameters.getSRPIdentity())
                         // TODO Consider filtering extensions that aren't relevant to resumed sessions
                         .setServerExtensions(this.serverExtensions)
                         .build();
@@ -181,7 +234,7 @@ public abstract class TlsProtocol
                     this.tlsSession = new TlsSessionImpl(this.tlsSession.getSessionID(), this.sessionParameters);
                 }
 
-                getContext().setResumableSession(this.tlsSession);
+                getContextAdmin().setResumableSession(this.tlsSession);
             }
 
             getPeer().notifyHandshakeComplete();
@@ -236,6 +289,7 @@ public abstract class TlsProtocol
             // TODO[RFC 6520]
 //            heartbeatQueue.addData(buf, offset, len);
 //            processHeartbeat();
+            break;
         }
         default:
             /*
@@ -243,6 +297,7 @@ public abstract class TlsProtocol
              * 
              * RFC2246 defines on page 13, that we should ignore this.
              */
+            break;
         }
     }
 
@@ -256,24 +311,25 @@ public abstract class TlsProtocol
             /*
              * We need the first 4 bytes, they contain type and length of the message.
              */
-            if (handshakeQueue.size() >= 4)
+            if (handshakeQueue.available() >= 4)
             {
                 byte[] beginning = new byte[4];
                 handshakeQueue.read(beginning, 0, 4, 0);
-                ByteArrayInputStream bis = new ByteArrayInputStream(beginning);
-                short type = TlsUtils.readUint8(bis);
-                int len = TlsUtils.readUint24(bis);
+                short type = TlsUtils.readUint8(beginning, 0);
+                int len = TlsUtils.readUint24(beginning, 1);
 
                 /*
                  * Check if we have enough bytes in the buffer to read the full message.
                  */
-                if (handshakeQueue.size() >= (len + 4))
+                if (handshakeQueue.available() >= (len + 4))
                 {
                     /*
                      * Read the message.
                      */
                     byte[] buf = handshakeQueue.removeData(len, 4);
 
+                    checkReceivedChangeCipherSpec(connection_state == CS_END || type == HandshakeType.finished);
+
                     /*
                      * RFC 2246 7.4.9. The value handshake_messages includes all handshake messages
                      * starting at client hello up to, but not including, this finished message.
@@ -285,9 +341,11 @@ public abstract class TlsProtocol
                         break;
                     case HandshakeType.finished:
                     {
-                        if (this.expected_verify_data == null)
+                        TlsContext ctx = getContext();
+                        if (this.expected_verify_data == null
+                            && ctx.getSecurityParameters().getMasterSecret() != null)
                         {
-                            this.expected_verify_data = createVerifyData(!getContext().isServer());
+                            this.expected_verify_data = createVerifyData(!ctx.isServer());
                         }
 
                         // NB: Fall through to next case label
@@ -321,7 +379,7 @@ public abstract class TlsProtocol
     private void processAlert()
         throws IOException
     {
-        while (alertQueue.size() >= 2)
+        while (alertQueue.available() >= 2)
         {
             /*
              * An alert is always 2 bytes. Read the alert.
@@ -387,8 +445,8 @@ public abstract class TlsProtocol
             }
 
             if (this.receivedChangeCipherSpec
-                || alertQueue.size() > 0
-                || handshakeQueue.size() > 0)
+                || alertQueue.available() > 0
+                || handshakeQueue.available() > 0)
             {
                 throw new TlsFatalAlert(AlertDescription.unexpected_message);
             }
@@ -402,9 +460,8 @@ public abstract class TlsProtocol
     }
 
     protected int applicationDataAvailable()
-        throws IOException
     {
-        return applicationDataQueue.size();
+        return applicationDataQueue.available();
     }
 
     /**
@@ -425,7 +482,7 @@ public abstract class TlsProtocol
             return 0;
         }
 
-        while (applicationDataQueue.size() == 0)
+        while (applicationDataQueue.available() == 0)
         {
             /*
              * We need to read some data.
@@ -449,7 +506,7 @@ public abstract class TlsProtocol
             safeReadRecord();
         }
 
-        len = Math.min(len, applicationDataQueue.size());
+        len = Math.min(len, applicationDataQueue.available());
         applicationDataQueue.removeData(buf, offset, len, 0);
         return len;
     }
@@ -468,7 +525,7 @@ public abstract class TlsProtocol
         }
         catch (TlsFatalAlert e)
         {
-            if (!this.closed)
+            if (!closed)
             {
                 this.failWithError(AlertLevel.fatal, e.getAlertDescription(), "Failed to read record", e);
             }
@@ -476,7 +533,7 @@ public abstract class TlsProtocol
         }
         catch (IOException e)
         {
-            if (!this.closed)
+            if (!closed)
             {
                 this.failWithError(AlertLevel.fatal, AlertDescription.internal_error, "Failed to read record", e);
             }
@@ -484,7 +541,7 @@ public abstract class TlsProtocol
         }
         catch (RuntimeException e)
         {
-            if (!this.closed)
+            if (!closed)
             {
                 this.failWithError(AlertLevel.fatal, AlertDescription.internal_error, "Failed to read record", e);
             }
@@ -501,7 +558,7 @@ public abstract class TlsProtocol
         }
         catch (TlsFatalAlert e)
         {
-            if (!this.closed)
+            if (!closed)
             {
                 this.failWithError(AlertLevel.fatal, e.getAlertDescription(), "Failed to write record", e);
             }
@@ -527,9 +584,9 @@ public abstract class TlsProtocol
 
     /**
      * Send some application data to the remote system.
-     * <p/>
+     * <p>
      * The method will handle fragmentation internally.
-     *
+     * </p>
      * @param buf    The buffer with the data.
      * @param offset The position in the buffer where the data is placed.
      * @param len    The length of the data.
@@ -557,16 +614,27 @@ public abstract class TlsProtocol
              * NOTE: Actually, implementations appear to have settled on 1/n-1 record splitting.
              */
 
-            if (this.splitApplicationDataRecords)
+            if (this.appDataSplitEnabled)
             {
                 /*
                  * Protect against known IV attack!
                  * 
                  * DO NOT REMOVE THIS CODE, EXCEPT YOU KNOW EXACTLY WHAT YOU ARE DOING HERE.
                  */
-                safeWriteRecord(ContentType.application_data, buf, offset, 1);
-                ++offset;
-                --len;
+                switch (appDataSplitMode) {
+                    case ADS_MODE_0_N_FIRSTONLY:
+                        this.appDataSplitEnabled = false;
+                        // fall through intended!
+                    case ADS_MODE_0_N:
+                        safeWriteRecord(ContentType.application_data, TlsUtils.EMPTY_BYTES, 0, 0);
+                        break;
+                    case ADS_MODE_1_Nsub1:
+                    default:
+                        safeWriteRecord(ContentType.application_data, buf, offset, 1);
+                        ++offset;
+                        --len;
+                        break;
+                }
             }
 
             if (len > 0)
@@ -580,6 +648,15 @@ public abstract class TlsProtocol
         }
     }
 
+    protected void setAppDataSplitMode(int appDataSplitMode) {
+        if (appDataSplitMode < ADS_MODE_1_Nsub1 ||
+            appDataSplitMode > ADS_MODE_0_N_FIRSTONLY)
+        {
+            throw new IllegalArgumentException("Illegal appDataSplitMode mode: " + appDataSplitMode);
+        }
+        this.appDataSplitMode = appDataSplitMode;
+	}
+
     protected void writeHandshakeMessage(byte[] buf, int off, int len) throws IOException
     {
         while (len > 0)
@@ -593,21 +670,198 @@ public abstract class TlsProtocol
     }
 
     /**
-     * @return An OutputStream which can be used to send data.
+     * @return An OutputStream which can be used to send data. Only allowed in blocking mode.
      */
     public OutputStream getOutputStream()
     {
+        if (!blocking)
+        {
+            throw new IllegalStateException("Cannot use OutputStream in non-blocking mode! Use offerOutput() instead.");
+        }
         return this.tlsOutputStream;
     }
 
     /**
-     * @return An InputStream which can be used to read data.
+     * @return An InputStream which can be used to read data. Only allowed in blocking mode.
      */
     public InputStream getInputStream()
     {
+        if (!blocking)
+        {
+            throw new IllegalStateException("Cannot use InputStream in non-blocking mode! Use offerInput() instead.");
+        }
         return this.tlsInputStream;
     }
 
+    /**
+     * Offer input from an arbitrary source. Only allowed in non-blocking mode.<br>
+     * <br>
+     * After this method returns, the input buffer is "owned" by this object. Other code
+     * must not attempt to do anything with it.<br>
+     * <br>
+     * This method will decrypt and process all records that are fully available.
+     * If only part of a record is available, the buffer will be retained until the
+     * remainder of the record is offered.<br>
+     * <br>
+     * If any records containing application data were processed, the decrypted data
+     * can be obtained using {@link #readInput(byte[], int, int)}. If any records
+     * containing protocol data were processed, a response may have been generated.
+     * You should always check to see if there is any available output after calling
+     * this method by calling {@link #getAvailableOutputBytes()}.
+     * @param input The input buffer to offer
+     * @throws IOException If an error occurs while decrypting or processing a record
+     */
+    public void offerInput(byte[] input) throws IOException
+    {
+        if (blocking)
+        {
+            throw new IllegalStateException("Cannot use offerInput() in blocking mode! Use getInputStream() instead.");
+        }
+        
+        if (closed)
+        {
+            throw new IOException("Connection is closed, cannot accept any more input");
+        }
+        
+        inputBuffers.addBytes(input);
+
+        // loop while there are enough bytes to read the length of the next record
+        while (inputBuffers.available() >= RecordStream.TLS_HEADER_SIZE)
+        {
+            byte[] header = new byte[RecordStream.TLS_HEADER_SIZE];
+            inputBuffers.peek(header);
+
+            int totalLength = TlsUtils.readUint16(header, RecordStream.TLS_HEADER_LENGTH_OFFSET) + RecordStream.TLS_HEADER_SIZE;
+            if (inputBuffers.available() < totalLength)
+            {
+                // not enough bytes to read a whole record
+                break;
+            }
+
+            safeReadRecord();
+        }
+    }
+
+    /**
+     * Gets the amount of received application data. A call to {@link #readInput(byte[], int, int)}
+     * is guaranteed to be able to return at least this much data.<br>
+     * <br>
+     * Only allowed in non-blocking mode.
+     * @return The number of bytes of available application data
+     */
+    public int getAvailableInputBytes()
+    {
+        if (blocking)
+        {
+            throw new IllegalStateException("Cannot use getAvailableInputBytes() in blocking mode! Use getInputStream().available() instead.");
+        }
+        return applicationDataAvailable();
+    }
+
+    /**
+     * Retrieves received application data. Use {@link #getAvailableInputBytes()} to check
+     * how much application data is currently available. This method functions similarly to
+     * {@link InputStream#read(byte[], int, int)}, except that it never blocks. If no data
+     * is available, nothing will be copied and zero will be returned.<br>
+     * <br>
+     * Only allowed in non-blocking mode.
+     * @param buffer The buffer to hold the application data
+     * @param offset The start offset in the buffer at which the data is written
+     * @param length The maximum number of bytes to read
+     * @return The total number of bytes copied to the buffer. May be less than the
+     *          length specified if the length was greater than the amount of available data.
+     */
+    public int readInput(byte[] buffer, int offset, int length)
+    {
+        if (blocking)
+        {
+            throw new IllegalStateException("Cannot use readInput() in blocking mode! Use getInputStream() instead.");
+        }
+        
+        try
+        {
+            return readApplicationData(buffer, offset, Math.min(length, applicationDataAvailable()));
+        }
+        catch (IOException e)
+        {
+            // readApplicationData() only throws if there is no data available, so this should never happen
+            throw new RuntimeException(e.toString()); // early JDK fix.
+        }
+    }
+
+    /**
+     * Offer output from an arbitrary source. Only allowed in non-blocking mode.<br>
+     * <br>
+     * After this method returns, the specified section of the buffer will have been
+     * processed. Use {@link #readOutput(byte[], int, int)} to get the bytes to
+     * transmit to the other peer.<br>
+     * <br>
+     * This method must not be called until after the handshake is complete! Attempting
+     * to call it before the handshake is complete will result in an exception.
+     * @param buffer The buffer containing application data to encrypt
+     * @param offset The offset at which to begin reading data
+     * @param length The number of bytes of data to read
+     * @throws IOException If an error occurs encrypting the data, or the handshake is not complete
+     */
+    public void offerOutput(byte[] buffer, int offset, int length)
+            throws IOException
+    {
+        if (blocking)
+        {
+            throw new IllegalStateException("Cannot use offerOutput() in blocking mode! Use getOutputStream() instead.");
+        }
+        
+        if (!appDataReady)
+        {
+            throw new IOException("Application data cannot be sent until the handshake is complete!");
+        }
+        
+        writeData(buffer, offset, length);
+    }
+
+    /**
+     * Gets the amount of encrypted data available to be sent. A call to
+     * {@link #readOutput(byte[], int, int)} is guaranteed to be able to return at
+     * least this much data.<br>
+     * <br>
+     * Only allowed in non-blocking mode.
+     * @return The number of bytes of available encrypted data
+     */
+    public int getAvailableOutputBytes()
+    {
+        if (blocking)
+        {
+            throw new IllegalStateException("Cannot use getAvailableOutputBytes() in blocking mode! Use getOutputStream() instead.");
+        }
+        
+        return outputBuffer.getBuffer().available();
+    }
+
+    /**
+     * Retrieves encrypted data to be sent. Use {@link #getAvailableOutputBytes()} to check
+     * how much encrypted data is currently available. This method functions similarly to
+     * {@link InputStream#read(byte[], int, int)}, except that it never blocks. If no data
+     * is available, nothing will be copied and zero will be returned.<br>
+     * <br>
+     * Only allowed in non-blocking mode.
+     * @param buffer The buffer to hold the encrypted data
+     * @param offset The start offset in the buffer at which the data is written
+     * @param length The maximum number of bytes to read
+     * @return The total number of bytes copied to the buffer. May be less than the
+     *          length specified if the length was greater than the amount of available data.
+     */
+    public int readOutput(byte[] buffer, int offset, int length)
+    {
+        if (blocking)
+        {
+            throw new IllegalStateException("Cannot use readOutput() in blocking mode! Use getOutputStream() instead.");
+        }
+        
+        int bytesToRead = Math.min(getAvailableOutputBytes(), length);
+        outputBuffer.getBuffer().removeData(buffer, offset, bytesToRead, 0);
+        return bytesToRead;
+    }
+
     /**
      * Terminate this connection with an alert. Can be used for normal closure too.
      * 
@@ -618,7 +872,7 @@ public abstract class TlsProtocol
      * @throws IOException
      *             If alert was fatal.
      */
-    protected void failWithError(short alertLevel, short alertDescription, String message, Exception cause)
+    protected void failWithError(short alertLevel, short alertDescription, String message, Throwable cause)
         throws IOException
     {
         /*
@@ -671,6 +925,11 @@ public abstract class TlsProtocol
     protected void processFinishedMessage(ByteArrayInputStream buf)
         throws IOException
     {
+        if (expected_verify_data == null)
+        {
+            throw new TlsFatalAlert(AlertDescription.internal_error);
+        }
+
         byte[] verify_data = TlsUtils.readFully(expected_verify_data.length, buf);
 
         assertEmpty(buf);
@@ -687,7 +946,7 @@ public abstract class TlsProtocol
         }
     }
 
-    protected void raiseAlert(short alertLevel, short alertDescription, String message, Exception cause)
+    protected void raiseAlert(short alertLevel, short alertDescription, String message, Throwable cause)
         throws IOException
     {
         getPeer().notifyAlertRaised(alertLevel, alertDescription, message, cause);
@@ -713,7 +972,7 @@ public abstract class TlsProtocol
             certificate = Certificate.EMPTY_CHAIN;
         }
 
-        if (certificate.getLength() == 0)
+        if (certificate.isEmpty())
         {
             TlsContext context = getContext();
             if (!context.isServer())
@@ -721,8 +980,8 @@ public abstract class TlsProtocol
                 ProtocolVersion serverVersion = getContext().getServerVersion();
                 if (serverVersion.isSSL())
                 {
-                    String message = serverVersion.toString() + " client didn't provide credentials";
-                    raiseWarning(AlertDescription.no_certificate, message);
+                    String errorMessage = serverVersion.toString() + " client didn't provide credentials";
+                    raiseWarning(AlertDescription.no_certificate, errorMessage);
                     return;
                 }
             }
@@ -768,15 +1027,10 @@ public abstract class TlsProtocol
     protected byte[] createVerifyData(boolean isServer)
     {
         TlsContext context = getContext();
-
-        if (isServer)
-        {
-            return TlsUtils.calculateVerifyData(context, ExporterLabel.server_finished,
-                getCurrentPRFHash(getContext(), recordStream.getHandshakeHash(), TlsUtils.SSL_SERVER));
-        }
-
-        return TlsUtils.calculateVerifyData(context, ExporterLabel.client_finished,
-            getCurrentPRFHash(getContext(), recordStream.getHandshakeHash(), TlsUtils.SSL_CLIENT));
+        String asciiLabel = isServer ? ExporterLabel.server_finished : ExporterLabel.client_finished;
+        byte[] sslSender = isServer ? TlsUtils.SSL_SERVER : TlsUtils.SSL_CLIENT;
+        byte[] hash = getCurrentPRFHash(context, recordStream.getHandshakeHash(), sslSender);
+        return TlsUtils.calculateVerifyData(context, asciiLabel, hash);
     }
 
     /**
@@ -809,13 +1063,21 @@ public abstract class TlsProtocol
         recordStream.flush();
     }
 
-    protected short processMaxFragmentLengthExtension(Hashtable clientExtensions, Hashtable serverExtensions, short alertDescription)
+    public boolean isClosed()
+    {
+        return closed;
+    }
+
+    protected short processMaxFragmentLengthExtension(Hashtable clientExtensions, Hashtable serverExtensions,
+        short alertDescription)
         throws IOException
     {
         short maxFragmentLength = TlsExtensionsUtils.getMaxFragmentLengthExtension(serverExtensions);
-        if (maxFragmentLength >= 0 && !this.resumedSession)
+        if (maxFragmentLength >= 0)
         {
-            if (maxFragmentLength != TlsExtensionsUtils.getMaxFragmentLengthExtension(clientExtensions))
+            if (!MaxFragmentLength.isValid(maxFragmentLength)
+                || (!this.resumedSession && maxFragmentLength != TlsExtensionsUtils
+                    .getMaxFragmentLengthExtension(clientExtensions)))
             {
                 throw new TlsFatalAlert(alertDescription);
             }
@@ -823,6 +1085,20 @@ public abstract class TlsProtocol
         return maxFragmentLength;
     }
 
+    protected void refuseRenegotiation() throws IOException
+    {
+        /*
+         * RFC 5746 4.5 SSLv3 clients that refuse renegotiation SHOULD use a fatal
+         * handshake_failure alert.
+         */
+        if (TlsUtils.isSSL(getContext()))
+        {
+            throw new TlsFatalAlert(AlertDescription.handshake_failure);
+        }
+
+        raiseWarning(AlertDescription.no_renegotiation, "Renegotiation not supported");
+    }
+
     /**
      * Make sure the InputStream 'buf' now empty. Fail otherwise.
      *
@@ -958,6 +1234,21 @@ public abstract class TlsProtocol
     {
         ByteArrayOutputStream buf = new ByteArrayOutputStream();
 
+        /*
+         * NOTE: There are reports of servers that don't accept a zero-length extension as the last
+         * one, so we write out any zero-length ones first as a best-effort workaround.
+         */
+        writeSelectedExtensions(buf, extensions, true);
+        writeSelectedExtensions(buf, extensions, false);
+
+        byte[] extBytes = buf.toByteArray();
+
+        TlsUtils.writeOpaque16(extBytes, output);
+    }
+
+    protected static void writeSelectedExtensions(OutputStream output, Hashtable extensions, boolean selectEmpty)
+        throws IOException
+    {
         Enumeration keys = extensions.keys();
         while (keys.hasMoreElements())
         {
@@ -965,14 +1256,13 @@ public abstract class TlsProtocol
             int extension_type = key.intValue();
             byte[] extension_data = (byte[])extensions.get(key);
 
-            TlsUtils.checkUint16(extension_type);
-            TlsUtils.writeUint16(extension_type, buf);
-            TlsUtils.writeOpaque16(extension_data, buf);
+            if (selectEmpty == (extension_data.length == 0))
+            {
+                TlsUtils.checkUint16(extension_type);
+                TlsUtils.writeUint16(extension_type, output);
+                TlsUtils.writeOpaque16(extension_data, output);
+            }
         }
-
-        byte[] extBytes = buf.toByteArray();
-
-        TlsUtils.writeOpaque16(extBytes, output);
     }
 
     protected static void writeSupplementalData(OutputStream output, Vector supplementalData)
@@ -1024,19 +1314,24 @@ public abstract class TlsProtocol
         case CipherSuite.TLS_DHE_DSS_WITH_CAMELLIA_256_CBC_SHA256:
         case CipherSuite.TLS_DHE_PSK_WITH_AES_128_CCM:
         case CipherSuite.TLS_DHE_PSK_WITH_AES_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_DHE_PSK_WITH_AES_128_OCB:
         case CipherSuite.TLS_DHE_PSK_WITH_AES_256_CCM:
+        case CipherSuite.DRAFT_TLS_DHE_PSK_WITH_AES_256_OCB:
         case CipherSuite.TLS_DHE_PSK_WITH_CAMELLIA_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_DHE_PSK_WITH_CHACHA20_POLY1305_SHA256:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_128_CBC_SHA256:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_128_CCM:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_128_CCM_8:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_DHE_RSA_WITH_AES_128_OCB:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_256_CBC_SHA256:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_256_CCM:
         case CipherSuite.TLS_DHE_RSA_WITH_AES_256_CCM_8:
+        case CipherSuite.DRAFT_TLS_DHE_RSA_WITH_AES_256_OCB:
         case CipherSuite.TLS_DHE_RSA_WITH_CAMELLIA_128_CBC_SHA256:
         case CipherSuite.TLS_DHE_RSA_WITH_CAMELLIA_128_GCM_SHA256:
         case CipherSuite.TLS_DHE_RSA_WITH_CAMELLIA_256_CBC_SHA256:
-        case CipherSuite.TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256:
+        case CipherSuite.DRAFT_TLS_DHE_RSA_WITH_CHACHA20_POLY1305_SHA256:
         case CipherSuite.TLS_ECDH_ECDSA_WITH_AES_128_CBC_SHA256:
         case CipherSuite.TLS_ECDH_ECDSA_WITH_AES_128_GCM_SHA256:
         case CipherSuite.TLS_ECDH_ECDSA_WITH_CAMELLIA_128_CBC_SHA256:
@@ -1046,25 +1341,40 @@ public abstract class TlsProtocol
         case CipherSuite.TLS_ECDH_RSA_WITH_CAMELLIA_128_CBC_SHA256:
         case CipherSuite.TLS_ECDH_RSA_WITH_CAMELLIA_128_GCM_SHA256:
         case CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CBC_SHA256:
+        case CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CCM:
+        case CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_CCM_8:
         case CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_ECDHE_ECDSA_WITH_AES_128_OCB:
+        case CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CCM:
+        case CipherSuite.TLS_ECDHE_ECDSA_WITH_AES_256_CCM_8:
+        case CipherSuite.DRAFT_TLS_ECDHE_ECDSA_WITH_AES_256_OCB:
         case CipherSuite.TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_CBC_SHA256:
         case CipherSuite.TLS_ECDHE_ECDSA_WITH_CAMELLIA_128_GCM_SHA256:
-        case CipherSuite.TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256:
+        case CipherSuite.DRAFT_TLS_ECDHE_ECDSA_WITH_CHACHA20_POLY1305_SHA256:
+        case CipherSuite.DRAFT_TLS_ECDHE_PSK_WITH_AES_128_OCB:
+        case CipherSuite.DRAFT_TLS_ECDHE_PSK_WITH_AES_256_OCB:
+        case CipherSuite.DRAFT_TLS_ECDHE_PSK_WITH_CHACHA20_POLY1305_SHA256:
         case CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_CBC_SHA256:
         case CipherSuite.TLS_ECDHE_RSA_WITH_AES_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_ECDHE_RSA_WITH_AES_128_OCB:
+        case CipherSuite.DRAFT_TLS_ECDHE_RSA_WITH_AES_256_OCB:
         case CipherSuite.TLS_ECDHE_RSA_WITH_CAMELLIA_128_CBC_SHA256:
         case CipherSuite.TLS_ECDHE_RSA_WITH_CAMELLIA_128_GCM_SHA256:
-        case CipherSuite.TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256:
+        case CipherSuite.DRAFT_TLS_ECDHE_RSA_WITH_CHACHA20_POLY1305_SHA256:
         case CipherSuite.TLS_PSK_DHE_WITH_AES_128_CCM_8:
         case CipherSuite.TLS_PSK_DHE_WITH_AES_256_CCM_8:
         case CipherSuite.TLS_PSK_WITH_AES_128_CCM:
         case CipherSuite.TLS_PSK_WITH_AES_128_CCM_8:
         case CipherSuite.TLS_PSK_WITH_AES_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_PSK_WITH_CHACHA20_POLY1305_SHA256:
+        case CipherSuite.DRAFT_TLS_PSK_WITH_AES_128_OCB:
         case CipherSuite.TLS_PSK_WITH_AES_256_CCM:
         case CipherSuite.TLS_PSK_WITH_AES_256_CCM_8:
+        case CipherSuite.DRAFT_TLS_PSK_WITH_AES_256_OCB:
         case CipherSuite.TLS_PSK_WITH_CAMELLIA_128_GCM_SHA256:
         case CipherSuite.TLS_RSA_PSK_WITH_AES_128_GCM_SHA256:
         case CipherSuite.TLS_RSA_PSK_WITH_CAMELLIA_128_GCM_SHA256:
+        case CipherSuite.DRAFT_TLS_RSA_PSK_WITH_CHACHA20_POLY1305_SHA256:
         case CipherSuite.TLS_RSA_WITH_AES_128_CBC_SHA256:
         case CipherSuite.TLS_RSA_WITH_AES_128_CCM:
         case CipherSuite.TLS_RSA_WITH_AES_128_CCM_8:
