@@ -181,19 +181,19 @@ public abstract class ECPoint
         int zsLen = zs.length;
         if (zsLen == 0)
         {
-            return zs;
+            return EMPTY_ZS;
         }
         ECFieldElement[] copy = new ECFieldElement[zsLen];
         System.arraycopy(zs, 0, copy, 0, zsLen);
         return copy;
     }
 
-    protected final ECFieldElement getRawXCoord()
+    public final ECFieldElement getRawXCoord()
     {
         return x;
     }
 
-    protected final ECFieldElement getRawYCoord()
+    public final ECFieldElement getRawYCoord()
     {
         return y;
     }
@@ -287,6 +287,9 @@ public abstract class ECPoint
         return x == null || y == null || (zs.length > 0 && zs[0].isZero());
     }
 
+    /**
+     * @deprecated per-point compression property will be removed, refer {@link #getEncoded(boolean)}
+     */
     public boolean isCompressed()
     {
         return this.withCompression;
@@ -435,13 +438,19 @@ public abstract class ECPoint
         return sb.toString();
     }
 
+    /**
+     * @deprecated per-point compression property will be removed, refer {@link #getEncoded(boolean)}
+     */
     public byte[] getEncoded()
     {
         return getEncoded(this.withCompression);
     }
 
     /**
-     * return the field element encoded with point compression. (S 4.3.6)
+     * Get an encoding of the point value, optionally in compressed format.
+     * 
+     * @param compressed whether to generate a compressed point encoding.
+     * @return the point encoding
      */
     public byte[] getEncoded(boolean compressed)
     {
@@ -771,14 +780,7 @@ public abstract class ECPoint
                     Y3 = W1.subtract(X3).multiply(dy).subtract(A1);
                     Z3 = dx;
 
-                    if (Z1IsOne)
-                    {
-                        Z3Squared = C;
-                    }
-                    else
-                    {
-                        Z3 = Z3.multiply(Z1);
-                    }
+                    Z3 = Z3.multiply(Z1);
                 }
                 else
                 {
@@ -967,7 +969,7 @@ public abstract class ECPoint
                     }
                     else if (!a4.isZero())
                     {
-                        ECFieldElement Z1Squared = Z1IsOne ? Z1 : Z1.square();
+                        ECFieldElement Z1Squared = Z1.square();
                         ECFieldElement Z1Pow4 = Z1Squared.square();
                         if (a4Neg.bitLength() < a4.bitLength())
                         {
@@ -1173,6 +1175,8 @@ public abstract class ECPoint
             {
                 switch (coord)
                 {
+                case ECCurve.COORD_AFFINE:
+                    break;
                 case ECCurve.COORD_HOMOGENEOUS:
                     ECFieldElement Z1Sq = Z1.square();
                     X1 = X1.multiply(Z1);
@@ -1185,6 +1189,8 @@ public abstract class ECPoint
                 case ECCurve.COORD_JACOBIAN_MODIFIED:
                     W1 = getJacobianModifiedW();
                     break;
+                default:
+                    throw new IllegalStateException("unsupported coordinate system");
                 }
             }
 
@@ -1416,6 +1422,153 @@ public abstract class ECPoint
             ECFieldElement rhs = X.add(A).multiply(X.square()).add(B);
             return lhs.equals(rhs);
         }
+
+        public ECPoint scaleX(ECFieldElement scale)
+        {
+            if (this.isInfinity())
+            {
+                return this;
+            }
+
+            int coord = this.getCurveCoordinateSystem();
+
+            switch (coord)
+            {
+            case ECCurve.COORD_LAMBDA_AFFINE:
+            {
+                // Y is actually Lambda (X + Y/X) here
+                ECFieldElement X = this.getRawXCoord(), L = this.getRawYCoord(); // earlier JDK
+
+                ECFieldElement X2 = X.multiply(scale);
+                ECFieldElement L2 = L.add(X).divide(scale).add(X2);
+
+                return this.getCurve().createRawPoint(X, L2, this.getRawZCoords(), this.withCompression); // earlier JDK
+            }
+            case ECCurve.COORD_LAMBDA_PROJECTIVE:
+            {
+                // Y is actually Lambda (X + Y/X) here
+                ECFieldElement X = this.getRawXCoord(), L = this.getRawYCoord(), Z = this.getRawZCoords()[0]; // earlier JDK
+
+                // We scale the Z coordinate also, to avoid an inversion
+                ECFieldElement X2 = X.multiply(scale.square());
+                ECFieldElement L2 = L.add(X).add(X2);
+                ECFieldElement Z2 = Z.multiply(scale);
+
+                return this.getCurve().createRawPoint(X2, L2, new ECFieldElement[]{ Z2 }, this.withCompression); // earlier JDK
+            }
+            default:
+            {
+                return super.scaleX(scale);
+            }
+            }
+        }
+
+        public ECPoint scaleY(ECFieldElement scale)
+        {
+            if (this.isInfinity())
+            {
+                return this;
+            }
+
+            int coord = this.getCurveCoordinateSystem();
+
+            switch (coord)
+            {
+            case ECCurve.COORD_LAMBDA_AFFINE:
+            case ECCurve.COORD_LAMBDA_PROJECTIVE:
+            {
+                ECFieldElement X = this.getRawXCoord(), L = this.getRawYCoord(); // earlier JDK
+
+                // Y is actually Lambda (X + Y/X) here
+                ECFieldElement L2 = L.add(X).multiply(scale).add(X);
+
+                return this.getCurve().createRawPoint(X, L2, this.getRawZCoords(), this.withCompression); // earlier JDK
+            }
+            default:
+            {
+                return super.scaleY(scale);
+            }
+            }
+        }
+
+        public ECPoint subtract(ECPoint b)
+        {
+            if (b.isInfinity())
+            {
+                return this;
+            }
+
+            // Add -b
+            return this.add(b.negate());
+        }
+
+        public ECPoint.AbstractF2m tau()
+        {
+            if (this.isInfinity())
+            {
+                return this;
+            }
+
+            ECCurve curve = this.getCurve();
+            int coord = curve.getCoordinateSystem();
+
+            ECFieldElement X1 = this.x;
+
+            switch (coord)
+            {
+            case ECCurve.COORD_AFFINE:
+            case ECCurve.COORD_LAMBDA_AFFINE:
+            {
+                ECFieldElement Y1 = this.y;
+                return (ECPoint.AbstractF2m)curve.createRawPoint(X1.square(), Y1.square(), this.withCompression);
+            }
+            case ECCurve.COORD_HOMOGENEOUS:
+            case ECCurve.COORD_LAMBDA_PROJECTIVE:
+            {
+                ECFieldElement Y1 = this.y, Z1 = this.zs[0];
+                return (ECPoint.AbstractF2m)curve.createRawPoint(X1.square(), Y1.square(),
+                    new ECFieldElement[]{ Z1.square() }, this.withCompression);
+            }
+            default:
+            {
+                throw new IllegalStateException("unsupported coordinate system");
+            }
+            }
+        }
+
+        public ECPoint.AbstractF2m tauPow(int pow)
+        {
+            if (this.isInfinity())
+            {
+                return this;
+            }
+
+            ECCurve curve = this.getCurve();
+            int coord = curve.getCoordinateSystem();
+
+            ECFieldElement X1 = this.x;
+
+            switch (coord)
+            {
+            case ECCurve.COORD_AFFINE:
+            case ECCurve.COORD_LAMBDA_AFFINE:
+            {
+                ECFieldElement Y1 = this.y;
+                return (ECPoint.AbstractF2m)curve.createRawPoint(X1.squarePow(pow), Y1.squarePow(pow), this.withCompression);
+            }
+            case ECCurve.COORD_HOMOGENEOUS:
+            case ECCurve.COORD_LAMBDA_PROJECTIVE:
+            {
+                ECFieldElement Y1 = this.y, Z1 = this.zs[0];
+                return (ECPoint.AbstractF2m)curve.createRawPoint(X1.squarePow(pow), Y1.squarePow(pow),
+                    new ECFieldElement[]{ Z1.squarePow(pow) }, this.withCompression);
+            }
+            default:
+            {
+                throw new IllegalStateException("unsupported coordinate system");
+            }
+            }
+        }
     }
 
     /**
@@ -1518,74 +1671,6 @@ public abstract class ECPoint
             }
         }
 
-        public ECPoint scaleX(ECFieldElement scale)
-        {
-            if (this.isInfinity())
-            {
-                return this;
-            }
-
-            int coord = this.getCurveCoordinateSystem();
-
-            switch (coord)
-            {
-            case ECCurve.COORD_LAMBDA_AFFINE:
-            {
-                // Y is actually Lambda (X + Y/X) here
-                ECFieldElement X = this.getRawXCoord(), L = this.getRawYCoord(); // earlier JDK
-
-                ECFieldElement X2 = X.multiply(scale);
-                ECFieldElement L2 = L.add(X).divide(scale).add(X2);
-
-                return this.getCurve().createRawPoint(X, L2, this.getRawZCoords(), this.withCompression); // earlier JDK
-            }
-            case ECCurve.COORD_LAMBDA_PROJECTIVE:
-            {
-                // Y is actually Lambda (X + Y/X) here
-                ECFieldElement X = this.getRawXCoord(), L = this.getRawYCoord(), Z = this.getRawZCoords()[0]; // earlier JDK
-
-                // We scale the Z coordinate also, to avoid an inversion
-                ECFieldElement X2 = X.multiply(scale.square());
-                ECFieldElement L2 = L.add(X).add(X2);
-                ECFieldElement Z2 = Z.multiply(scale);
-
-                return this.getCurve().createRawPoint(X2, L2, new ECFieldElement[]{ Z2 }, this.withCompression); // earlier JDK
-            }
-            default:
-            {
-                return super.scaleX(scale);
-            }
-            }
-        }
-
-        public ECPoint scaleY(ECFieldElement scale)
-        {
-            if (this.isInfinity())
-            {
-                return this;
-            }
-
-            int coord = this.getCurveCoordinateSystem();
-
-            switch (coord)
-            {
-            case ECCurve.COORD_LAMBDA_AFFINE:
-            case ECCurve.COORD_LAMBDA_PROJECTIVE:
-            {
-                ECFieldElement X = this.getRawXCoord(), L = this.getRawYCoord(); // earlier JDK
-
-                // Y is actually Lambda (X + Y/X) here
-                ECFieldElement L2 = L.add(X).multiply(scale).add(X);
-
-                return this.getCurve().createRawPoint(X, L2, this.getRawZCoords(), this.withCompression); // earlier JDK
-            }
-            default:
-            {
-                return super.scaleY(scale);
-            }
-            }
-        }
-
         protected boolean getCompressionYTilde()
         {
             ECFieldElement X = this.getRawXCoord();
@@ -1611,44 +1696,7 @@ public abstract class ECPoint
             }
         }
 
-        /**
-         * Check, if two <code>ECPoint</code>s can be added or subtracted.
-         * @param a The first <code>ECPoint</code> to check.
-         * @param b The second <code>ECPoint</code> to check.
-         * @throws IllegalArgumentException if <code>a</code> and <code>b</code>
-         * cannot be added.
-         */
-        private static void checkPoints(ECPoint a, ECPoint b)
-        {
-            // Check, if points are on the same curve
-            if (a.curve != b.curve)
-            {
-                throw new IllegalArgumentException("Only points on the same "
-                        + "curve can be added or subtracted");
-            }
-
-//            ECFieldElement.F2m.checkFieldElements(a.x, b.x);
-        }
-
-        /* (non-Javadoc)
-         * @see org.bouncycastle.math.ec.ECPoint#add(org.bouncycastle.math.ec.ECPoint)
-         */
         public ECPoint add(ECPoint b)
-        {
-            checkPoints(this, b);
-            return addSimple((ECPoint.F2m)b);
-        }
-
-        /**
-         * Adds another <code>ECPoints.F2m</code> to <code>this</code> without
-         * checking if both points are on the same curve. Used by multiplication
-         * algorithms, because there all points are a multiple of the same point
-         * and hence the checks can be omitted.
-         * @param b The other <code>ECPoints.F2m</code> to add to
-         * <code>this</code>.
-         * @return <code>this + b</code>
-         */
-        public ECPoint.F2m addSimple(ECPoint.F2m b)
         {
             if (this.isInfinity())
             {
@@ -1677,10 +1725,10 @@ public abstract class ECPoint
                 {
                     if (dy.isZero())
                     {
-                        return (ECPoint.F2m)twice();
+                        return twice();
                     }
 
-                    return (ECPoint.F2m)curve.getInfinity();
+                    return curve.getInfinity();
                 }
 
                 ECFieldElement L = dy.divide(dx);
@@ -1708,10 +1756,10 @@ public abstract class ECPoint
                 {
                     if (U.isZero())
                     {
-                        return (ECPoint.F2m)twice();
+                        return twice();
                     }
 
-                    return (ECPoint.F2m)curve.getInfinity();
+                    return curve.getInfinity();
                 }
 
                 ECFieldElement VSq = V.square();
@@ -1733,10 +1781,10 @@ public abstract class ECPoint
                 {
                     if (X2.isZero())
                     {
-                        return (ECPoint.F2m)curve.getInfinity();
+                        return curve.getInfinity();
                     }
 
-                    return b.addSimple(this);
+                    return b.add(this);
                 }
 
                 ECFieldElement L1 = this.y, Z1 = this.zs[0];
@@ -1765,10 +1813,10 @@ public abstract class ECPoint
                 {
                     if (A.isZero())
                     {
-                        return (ECPoint.F2m)twice();
+                        return twice();
                     }
 
-                    return (ECPoint.F2m)curve.getInfinity();
+                    return curve.getInfinity();
                 }
 
                 ECFieldElement X3, L3, Z3;
@@ -1829,68 +1877,6 @@ public abstract class ECPoint
             }
         }
 
-        /* (non-Javadoc)
-         * @see org.bouncycastle.math.ec.ECPoint#subtract(org.bouncycastle.math.ec.ECPoint)
-         */
-        public ECPoint subtract(ECPoint b)
-        {
-            checkPoints(this, b);
-            return subtractSimple((ECPoint.F2m)b);
-        }
-
-        /**
-         * Subtracts another <code>ECPoints.F2m</code> from <code>this</code>
-         * without checking if both points are on the same curve. Used by
-         * multiplication algorithms, because there all points are a multiple
-         * of the same point and hence the checks can be omitted.
-         * @param b The other <code>ECPoints.F2m</code> to subtract from
-         * <code>this</code>.
-         * @return <code>this - b</code>
-         */
-        public ECPoint.F2m subtractSimple(ECPoint.F2m b)
-        {
-            if (b.isInfinity())
-            {
-                return this;
-            }
-
-            // Add -b
-            return addSimple((ECPoint.F2m)b.negate());
-        }
-
-        public ECPoint.F2m tau()
-        {
-            if (this.isInfinity())
-            {
-                return this;
-            }
-
-            ECCurve curve = this.getCurve();
-            int coord = curve.getCoordinateSystem();
-
-            ECFieldElement X1 = this.x;
-
-            switch (coord)
-            {
-            case ECCurve.COORD_AFFINE:
-            case ECCurve.COORD_LAMBDA_AFFINE:
-            {
-                ECFieldElement Y1 = this.y;
-                return new ECPoint.F2m(curve, X1.square(), Y1.square(), this.withCompression);
-            }
-            case ECCurve.COORD_HOMOGENEOUS:
-            case ECCurve.COORD_LAMBDA_PROJECTIVE:
-            {
-                ECFieldElement Y1 = this.y, Z1 = this.zs[0];
-                return new ECPoint.F2m(curve, X1.square(), Y1.square(), new ECFieldElement[]{ Z1.square() }, this.withCompression);
-            }
-            default:
-            {
-                throw new IllegalStateException("unsupported coordinate system");
-            }
-            }
-        }
-
         public ECPoint twice()
         {
             if (this.isInfinity()) 
