@@ -4,10 +4,13 @@ import java.util.ArrayList;
 import java.util.Enumeration;
 import java.util.List;
 
+import junit.extensions.TestSetup;
 import junit.framework.Test;
+import junit.framework.TestCase;
 import junit.framework.TestSuite;
 
-public class AllTests 
+public class AllTests
+    extends TestCase
 {
     public static void main (String[] args) 
         throws Exception
@@ -20,10 +23,10 @@ public class AllTests
     {   
         TestSuite suite = new TestSuite("EC Math tests");
 
-        suite.addTest(ECAlgorithmsTest.suite());
-        suite.addTest(ECPointTest.suite());
+        suite.addTestSuite(ECAlgorithmsTest.class);
+        suite.addTestSuite(ECPointTest.class);
 
-        return suite;
+        return new BCTestSetup(suite);
     }
 
     static List enumToList(Enumeration en)
@@ -37,4 +40,23 @@ public class AllTests
 
         return rv;
     }
+
+    static class BCTestSetup
+        extends TestSetup
+    {
+        public BCTestSetup(Test test)
+        {
+            super(test);
+        }
+
+        protected void setUp()
+        {
+
+        }
+
+        protected void tearDown()
+        {
+
+        }
+    }
 }
