@@ -1,5 +1,6 @@
 package org.bouncycastle.pqc.crypto.test;
 
+import junit.extensions.TestSetup;
 import junit.framework.Test;
 import junit.framework.TestCase;
 import junit.framework.TestSuite;
@@ -7,22 +8,7 @@ import org.bouncycastle.util.test.SimpleTestResult;
 
 public class AllTests
     extends TestCase
-{   
-    public void testCrypto()
-    {   
-        org.bouncycastle.util.test.Test[] tests = RegressionTest.tests;
-        
-        for (int i = 0; i != tests.length; i++)
-        {
-            SimpleTestResult  result = (SimpleTestResult)tests[i].perform();
-
-            if (!result.isSuccessful())
-            {
-                fail(result.toString());
-            }
-        }
-    }
-    
+{
     public static void main (String[] args)
     {
         junit.textui.TestRunner.run(suite());
@@ -31,8 +17,7 @@ public class AllTests
     public static Test suite()
     {
         TestSuite suite = new TestSuite("Lightweight PQ Crypto Tests");
-        
-        suite.addTestSuite(AllTests.class);
+
         suite.addTestSuite(BitStringTest.class);
         suite.addTestSuite(EncryptionKeyTest.class);
         suite.addTestSuite(NTRUEncryptionParametersTest.class);
@@ -41,7 +26,50 @@ public class AllTests
         suite.addTestSuite(NTRUSignatureKeyTest.class);
         suite.addTestSuite(NTRUSignerTest.class);
         suite.addTestSuite(NTRUSigningParametersTest.class);
+        suite.addTestSuite(AllTests.SimpleTestTest.class);
 
-        return suite;
+        return new BCTestSetup(suite);
+    }
+
+    public static class SimpleTestTest
+       extends TestCase
+    {
+        public void testPQC()
+        {
+            org.bouncycastle.util.test.Test[] tests = RegressionTest.tests;
+
+            for (int i = 0; i != tests.length; i++)
+            {
+                SimpleTestResult  result = (SimpleTestResult)tests[i].perform();
+
+                if (!result.isSuccessful())
+                {
+                    if (result.getException() != null)
+                    {
+                        result.getException().printStackTrace();
+                    }
+                    fail(result.toString());
+                }
+            }
+        }
+    }
+
+    static class BCTestSetup
+        extends TestSetup
+    {
+        public BCTestSetup(Test test)
+        {
+            super(test);
+        }
+
+        protected void setUp()
+        {
+
+        }
+
+        protected void tearDown()
+        {
+
+        }
     }
 }
