@@ -49,8 +49,8 @@ import org.bouncycastle.asn1.cms.Time;
 import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
 import org.bouncycastle.asn1.x509.AuthorityKeyIdentifier;
 import org.bouncycastle.asn1.x509.ExtendedKeyUsage;
+import org.bouncycastle.asn1.x509.Extension;
 import org.bouncycastle.asn1.x509.KeyPurposeId;
-import org.bouncycastle.asn1.x509.X509Extensions;
 import org.bouncycastle.cert.jcajce.JcaCertStoreBuilder;
 import org.bouncycastle.cms.SignerInformation;
 import org.bouncycastle.cms.SignerInformationStore;
@@ -72,10 +72,10 @@ public class SignedMailValidator
 
     private static final Class DEFAULT_CERT_PATH_REVIEWER = PKIXCertPathReviewer.class;
 
-    private static final String EXT_KEY_USAGE = X509Extensions.ExtendedKeyUsage
+    private static final String EXT_KEY_USAGE = Extension.extendedKeyUsage
         .getId();
 
-    private static final String SUBJECT_ALTERNATIVE_NAME = X509Extensions.SubjectAlternativeName
+    private static final String SUBJECT_ALTERNATIVE_NAME = Extension.subjectAlternativeName
         .getId();
 
     private static final int shortKeyLength = 512;
@@ -194,14 +194,14 @@ public class SignedMailValidator
 
             int fromsLength = (froms != null) ? froms.length : 0;
             fromAddresses = new String[fromsLength + ((sender != null) ? 1 : 0)];
-            for (int i = 0; i < froms.length; i++)
+            for (int i = 0; i < fromsLength; i++)
             {
                 InternetAddress inetAddr = (InternetAddress)froms[i];
                 fromAddresses[i] = inetAddr.getAddress();
             }
             if (sender != null)
             {
-                fromAddresses[froms.length] = sender.getAddress();
+                fromAddresses[fromsLength] = sender.getAddress();
             }
 
             // initialize results
@@ -310,8 +310,12 @@ public class SignedMailValidator
                 {
                     ErrorBundle msg = new ErrorBundle(RESOURCE_NAME,
                         "SignedMailValidator.noSigningTime");
-                    errors.add(msg);
-                    signTime = new Date();
+                    notifications.add(msg);
+                    signTime = pkixParam.getDate();
+                    if (signTime == null)
+                    {
+                        signTime = new Date();
+                    }
                 }
                 else
                 {
@@ -744,7 +748,7 @@ public class SignedMailValidator
                 {
                     throw new IllegalStateException(e.toString());
                 }
-                byte[] authKeyIdentBytes = cert.getExtensionValue(X509Extensions.AuthorityKeyIdentifier.getId());
+                byte[] authKeyIdentBytes = cert.getExtensionValue(Extension.authorityKeyIdentifier.getId());
                 if (authKeyIdentBytes != null)
                 {
                     try
