@@ -5,12 +5,16 @@ import java.io.OutputStream;
 import java.util.Enumeration;
 import java.util.Hashtable;
 
+import org.bouncycastle.util.Strings;
+
 /**
  * Output stream that writes data in ASCII Armored format.
  */
 public class ArmoredOutputStream
     extends OutputStream
 {
+    public static final String VERSION_HDR = "Version";
+
     private static final byte[] encodingTable =
         {
             (byte)'A', (byte)'B', (byte)'C', (byte)'D', (byte)'E', (byte)'F', (byte)'G',
@@ -85,7 +89,7 @@ public class ArmoredOutputStream
     boolean         clearText = false;
     boolean         newLine = false;
 
-    String          nl = System.getProperty("line.separator");
+    String          nl = Strings.lineSeparator();
 
     String          type;
     String          headerStart = "-----BEGIN PGP ";
@@ -112,7 +116,7 @@ public class ArmoredOutputStream
             nl = "\r\n";
         }
 
-        resetHeaders();
+        headers.put(VERSION_HDR, version);
     }
 
     /**
@@ -139,7 +143,7 @@ public class ArmoredOutputStream
     }
 
     /**
-     * Set an additional header entry.
+     * Set an additional header entry. A null value will clear the entry for name.
      *
      * @param name the name of the header entry.
      * @param value the value of the header entry.
@@ -148,16 +152,29 @@ public class ArmoredOutputStream
         String name,
         String value)
     {
-        this.headers.put(name, value);
+        if (value == null)
+        {
+            this.headers.remove(name);
+        }
+        else
+        {
+            this.headers.put(name, value);
+        }
     }
 
     /**
-     * Reset the headers to only contain a Version string.
+     * Reset the headers to only contain a Version string (if one is present)
      */
     public void resetHeaders()
     {
+        String version = (String)headers.get(VERSION_HDR);
+
         headers.clear();
-        headers.put("Version", version);
+
+        if (version != null)
+        {
+            headers.put(VERSION_HDR, version);
+        }
     }
 
     /**
@@ -321,14 +338,17 @@ public class ArmoredOutputStream
                 out.write(nl.charAt(i));
             }
 
-            writeHeaderEntry("Version", (String)headers.get("Version"));
+            if (headers.containsKey(VERSION_HDR))
+            {
+                writeHeaderEntry(VERSION_HDR, (String)headers.get("Version"));
+            }
 
             Enumeration e = headers.keys();
             while (e.hasMoreElements())
             {
                 String  key = (String)e.nextElement();
 
-                if (!key.equals("Version"))
+                if (!key.equals(VERSION_HDR))
                 {
                     writeHeaderEntry(key, (String)headers.get(key));
                 }
@@ -365,7 +385,7 @@ public class ArmoredOutputStream
     }
 
     /**
-     * <b>Note</b>: close does nor close the underlying stream. So it is possible to write
+     * <b>Note</b>: close() does not close the underlying stream. So it is possible to write
      * multiple objects using armoring to a single stream.
      */
     public void close()
