@@ -6,15 +6,21 @@ import java.security.KeyPair;
 import java.security.PrivateKey;
 import java.security.Security;
 import java.security.cert.X509Certificate;
+import java.text.SimpleDateFormat;
 import java.util.ArrayList;
 import java.util.Collection;
 import java.util.Date;
 import java.util.HashSet;
 import java.util.List;
 import java.util.Map;
+import java.util.Set;
+import java.util.SimpleTimeZone;
 
 import junit.framework.TestCase;
+import org.bouncycastle.asn1.ASN1GeneralizedTime;
 import org.bouncycastle.asn1.ASN1ObjectIdentifier;
+import org.bouncycastle.asn1.DEROctetString;
+import org.bouncycastle.asn1.DERUTF8String;
 import org.bouncycastle.asn1.cmp.PKIFailureInfo;
 import org.bouncycastle.asn1.cmp.PKIStatus;
 import org.bouncycastle.asn1.cms.AttributeTable;
@@ -24,6 +30,9 @@ import org.bouncycastle.asn1.ess.SigningCertificate;
 import org.bouncycastle.asn1.ess.SigningCertificateV2;
 import org.bouncycastle.asn1.pkcs.PKCSObjectIdentifiers;
 import org.bouncycastle.asn1.x500.X500Name;
+import org.bouncycastle.asn1.x509.Extension;
+import org.bouncycastle.asn1.x509.Extensions;
+import org.bouncycastle.asn1.x509.ExtensionsGenerator;
 import org.bouncycastle.asn1.x509.GeneralName;
 import org.bouncycastle.asn1.x509.GeneralNames;
 import org.bouncycastle.asn1.x509.IssuerSerial;
@@ -66,53 +75,58 @@ public class NewTSPTest
     public void testGeneral()
         throws Exception
     {
-            String signDN = "O=Bouncy Castle, C=AU";
-            KeyPair signKP = TSPTestUtil.makeKeyPair();
-            X509Certificate signCert = TSPTestUtil.makeCACertificate(signKP,
-                    signDN, signKP, signDN);
+        String signDN = "O=Bouncy Castle, C=AU";
+        KeyPair signKP = TSPTestUtil.makeKeyPair();
+        X509Certificate signCert = TSPTestUtil.makeCACertificate(signKP,
+            signDN, signKP, signDN);
 
-            String origDN = "CN=Eric H. Echidna, E=eric@bouncycastle.org, O=Bouncy Castle, C=AU";
-            KeyPair origKP = TSPTestUtil.makeKeyPair();
-            X509Certificate origCert = TSPTestUtil.makeCertificate(origKP,
-                    origDN, signKP, signDN);
+        String origDN = "CN=Eric H. Echidna, E=eric@bouncycastle.org, O=Bouncy Castle, C=AU";
+        KeyPair origKP = TSPTestUtil.makeKeyPair();
+        X509Certificate origCert = TSPTestUtil.makeCertificate(origKP,
+            origDN, signKP, signDN);
 
 
-            
-            List certList = new ArrayList();
-            certList.add(origCert);
-            certList.add(signCert);
+        List certList = new ArrayList();
+        certList.add(origCert);
+        certList.add(signCert);
 
-            Store certs = new JcaCertStore(certList);
-            
-            basicTest(origKP.getPrivate(), origCert, certs);
-            basicSha256Test(origKP.getPrivate(), origCert, certs);
-            basicTestWithTSA(origKP.getPrivate(), origCert, certs);
-            overrideAttrsTest(origKP.getPrivate(), origCert, certs);
-            responseValidationTest(origKP.getPrivate(), origCert, certs);
-            incorrectHashTest(origKP.getPrivate(), origCert, certs);
-            badAlgorithmTest(origKP.getPrivate(), origCert, certs);
-            timeNotAvailableTest(origKP.getPrivate(), origCert, certs);
-            badPolicyTest(origKP.getPrivate(), origCert, certs);
-            tokenEncodingTest(origKP.getPrivate(), origCert, certs);
-            certReqTest(origKP.getPrivate(), origCert, certs);
-            testAccuracyZeroCerts(origKP.getPrivate(), origCert, certs);
-            testAccuracyWithCertsAndOrdering(origKP.getPrivate(), origCert, certs);
-            testNoNonse(origKP.getPrivate(), origCert, certs);
+        Store certs = new JcaCertStore(certList);
+
+        basicTest(origKP.getPrivate(), origCert, certs);
+        resolutionTest(origKP.getPrivate(), origCert, certs, TimeStampTokenGenerator.R_SECONDS, "19700101000009Z");
+        resolutionTest(origKP.getPrivate(), origCert, certs, TimeStampTokenGenerator.R_TENTHS_OF_SECONDS, "19700101000009.9Z");
+        resolutionTest(origKP.getPrivate(), origCert, certs, TimeStampTokenGenerator.R_MICROSECONDS, "19700101000009.99Z");
+        resolutionTest(origKP.getPrivate(), origCert, certs, TimeStampTokenGenerator.R_MILLISECONDS, "19700101000009.999Z");
+        basicSha256Test(origKP.getPrivate(), origCert, certs);
+        basicTestWithTSA(origKP.getPrivate(), origCert, certs);
+        overrideAttrsTest(origKP.getPrivate(), origCert, certs);
+        responseValidationTest(origKP.getPrivate(), origCert, certs);
+        incorrectHashTest(origKP.getPrivate(), origCert, certs);
+        badAlgorithmTest(origKP.getPrivate(), origCert, certs);
+        timeNotAvailableTest(origKP.getPrivate(), origCert, certs);
+        badPolicyTest(origKP.getPrivate(), origCert, certs);
+        tokenEncodingTest(origKP.getPrivate(), origCert, certs);
+        certReqTest(origKP.getPrivate(), origCert, certs);
+        testAccuracyZeroCerts(origKP.getPrivate(), origCert, certs);
+        testAccuracyWithCertsAndOrdering(origKP.getPrivate(), origCert, certs);
+        testNoNonse(origKP.getPrivate(), origCert, certs);
+        extensionTest(origKP.getPrivate(), origCert, certs);
+        additionalExtensionTest(origKP.getPrivate(), origCert, certs);
     }
-    
+
     private void basicTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
         Store certs)
         throws Exception
     {
         TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(
-                new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
-        
+            new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
+
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -120,28 +134,82 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
 
-        AttributeTable  table = tsToken.getSignedAttributes();
+        AttributeTable table = tsToken.getSignedAttributes();
 
         assertNotNull("no signingCertificate attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificate));
     }
 
+    private void resolutionTest(
+        PrivateKey privateKey,
+        X509Certificate cert,
+        Store certs,
+        int   resolution,
+        String timeString)
+        throws Exception
+    {
+        TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(
+            new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
+
+        tsTokenGen.addCertificates(certs);
+
+        tsTokenGen.setResolution(resolution);
+
+        TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+
+        TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
+
+        TimeStampResponse tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date(9999L));
+
+        tsResp = new TimeStampResponse(tsResp.getEncoded());
+
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
+
+        SimpleDateFormat dateF = new SimpleDateFormat("yyyyMMddHHmmss.SSS'Z'");
+
+        dateF.setTimeZone(new SimpleTimeZone(0, "Z"));
+
+        assertEquals(timeString, tsToken.getTimeStampInfo().toASN1Structure().getGenTime().getTimeString());
+
+        // test zero truncation
+        tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date(9000L));
+        tsToken = tsResp.getTimeStampToken();
+
+        assertEquals("19700101000009Z", tsToken.getTimeStampInfo().toASN1Structure().getGenTime().getTimeString());
+
+        if (resolution > TimeStampTokenGenerator.R_MICROSECONDS)
+        {
+            tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date(9990L));
+            tsToken = tsResp.getTimeStampToken();
+
+            assertEquals("19700101000009.99Z", tsToken.getTimeStampInfo().toASN1Structure().getGenTime().getTimeString());
+        }
+        if (resolution > TimeStampTokenGenerator.R_TENTHS_OF_SECONDS)
+        {
+            tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date(9900L));
+            tsToken = tsResp.getTimeStampToken();
+
+            assertEquals("19700101000009.9Z", tsToken.getTimeStampInfo().toASN1Structure().getGenTime().getTimeString());
+        }
+    }
+
     private void basicSha256Test(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
         Store certs)
         throws Exception
     {
         TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(
-                new JcaSimpleSignerInfoGeneratorBuilder().build("SHA256withRSA", privateKey, cert), new SHA256DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
+            new JcaSimpleSignerInfoGeneratorBuilder().build("SHA256withRSA", privateKey, cert), new SHA256DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
 
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA256, new byte[32], BigInteger.valueOf(100));
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA256, new byte[32], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -151,11 +219,11 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
 
-        AttributeTable  table = tsToken.getSignedAttributes();
+        AttributeTable table = tsToken.getSignedAttributes();
 
         assertNotNull("no signingCertificate attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificateV2));
 
@@ -175,7 +243,7 @@ public class NewTSPTest
     }
 
     private void overrideAttrsTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
         Store certs)
         throws Exception
@@ -227,7 +295,7 @@ public class NewTSPTest
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -235,11 +303,11 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
 
-        AttributeTable  table = tsToken.getSignedAttributes();
+        AttributeTable table = tsToken.getSignedAttributes();
 
         assertNotNull("no signingCertificate attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificate));
         assertNotNull("no signingCertificateV2 attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificateV2));
@@ -258,19 +326,19 @@ public class NewTSPTest
     }
 
     private void basicTestWithTSA(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
         Store certs)
         throws Exception
     {
         TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(
-                new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
+            new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
 
         tsTokenGen.addCertificates(certs);
         tsTokenGen.setTSA(new GeneralName(new X500Name("CN=Test")));
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -278,19 +346,58 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
 
-        AttributeTable  table = tsToken.getSignedAttributes();
+        AttributeTable table = tsToken.getSignedAttributes();
 
         assertNotNull("no signingCertificate attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificate));
     }
 
-    private void responseValidationTest(
-        PrivateKey      privateKey,
+    private void additionalExtensionTest(
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
+        throws Exception
+    {
+        TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(
+            new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
+
+        tsTokenGen.addCertificates(certs);
+        tsTokenGen.setTSA(new GeneralName(new X500Name("CN=Test")));
+
+        TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+
+        TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
+
+        ExtensionsGenerator extGen = new ExtensionsGenerator();
+
+        extGen.addExtension(Extension.auditIdentity, false, new DERUTF8String("Test"));
+
+        TimeStampResponse tsResp = tsRespGen.generateGrantedResponse(request, new BigInteger("23"), new Date(), "Okay", extGen.generate());
+
+        tsResp = new TimeStampResponse(tsResp.getEncoded());
+
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
+
+        tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
+
+        AttributeTable table = tsToken.getSignedAttributes();
+
+        assertNotNull("no signingCertificate attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificate));
+
+        Extensions ext = tsToken.getTimeStampInfo().getExtensions();
+
+        assertEquals(1, ext.getExtensionOIDs().length);
+        assertEquals(new Extension(Extension.auditIdentity, false, new DERUTF8String("Test").getEncoded()), ext.getExtension(Extension.auditIdentity));
+    }
+
+    private void responseValidationTest(
+        PrivateKey privateKey,
+        X509Certificate cert,
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -301,7 +408,7 @@ public class NewTSPTest
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -309,21 +416,21 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider("BC").build(cert));
-        
+
         //
         // check validation
         //
         tsResp.validate(request);
-        
+
         try
         {
             request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(101));
-            
+
             tsResp.validate(request);
-            
+
             fail("response validation failed on invalid nonce.");
         }
         catch (TSPValidationException e)
@@ -334,22 +441,22 @@ public class NewTSPTest
         try
         {
             request = reqGen.generate(TSPAlgorithms.SHA1, new byte[22], BigInteger.valueOf(100));
-            
+
             tsResp.validate(request);
-            
+
             fail("response validation failed on wrong digest.");
         }
         catch (TSPValidationException e)
         {
             // ignore
         }
-        
+
         try
         {
             request = reqGen.generate(TSPAlgorithms.MD5, new byte[20], BigInteger.valueOf(100));
-            
+
             tsResp.validate(request);
-            
+
             fail("response validation failed on wrong digest.");
         }
         catch (TSPValidationException e)
@@ -357,21 +464,21 @@ public class NewTSPTest
             // ignore
         }
     }
-    
+
     private void incorrectHashTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
 
         TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(infoGeneratorBuilder.build(new JcaContentSignerBuilder("SHA1withRSA").setProvider(BC).build(privateKey), cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
-        
+
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[16]);
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[16]);
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -379,30 +486,30 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         if (tsToken != null)
         {
             fail("incorrectHash - token not null.");
         }
-        
-        PKIFailureInfo  failInfo = tsResp.getFailInfo();
-        
+
+        PKIFailureInfo failInfo = tsResp.getFailInfo();
+
         if (failInfo == null)
         {
             fail("incorrectHash - failInfo set to null.");
         }
-        
+
         if (failInfo.intValue() != PKIFailureInfo.badDataFormat)
         {
             fail("incorrectHash - wrong failure info returned.");
         }
     }
-    
+
     private void badAlgorithmTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSimpleSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSimpleSignerInfoGeneratorBuilder().setProvider(BC);
@@ -412,7 +519,7 @@ public class NewTSPTest
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest            request = reqGen.generate(new ASN1ObjectIdentifier("1.2.3.4.5"), new byte[20]);
+        TimeStampRequest request = reqGen.generate(new ASN1ObjectIdentifier("1.2.3.4.5"), new byte[20]);
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -420,20 +527,20 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         if (tsToken != null)
         {
             fail("badAlgorithm - token not null.");
         }
 
-        PKIFailureInfo  failInfo = tsResp.getFailInfo();
-        
+        PKIFailureInfo failInfo = tsResp.getFailInfo();
+
         if (failInfo == null)
         {
             fail("badAlgorithm - failInfo set to null.");
         }
-        
+
         if (failInfo.intValue() != PKIFailureInfo.badAlg)
         {
             fail("badAlgorithm - wrong failure info returned.");
@@ -441,9 +548,9 @@ public class NewTSPTest
     }
 
     private void timeNotAvailableTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -453,7 +560,7 @@ public class NewTSPTest
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest            request = reqGen.generate(new ASN1ObjectIdentifier("1.2.3.4.5"), new byte[20]);
+        TimeStampRequest request = reqGen.generate(new ASN1ObjectIdentifier("1.2.3.4.5"), new byte[20]);
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -470,14 +577,14 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         if (tsToken != null)
         {
             fail("timeNotAvailable - token not null.");
         }
 
-        PKIFailureInfo  failInfo = tsResp.getFailInfo();
+        PKIFailureInfo failInfo = tsResp.getFailInfo();
 
         if (failInfo == null)
         {
@@ -491,9 +598,9 @@ public class NewTSPTest
     }
 
     private void badPolicyTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -503,10 +610,10 @@ public class NewTSPTest
         tsTokenGen.addCertificates(certs);
 
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        
+
         reqGen.setReqPolicy(new ASN1ObjectIdentifier("1.1"));
-        
-        TimeStampRequest            request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20]);
+
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20]);
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED, new HashSet());
 
@@ -523,30 +630,30 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         if (tsToken != null)
         {
             fail("badPolicy - token not null.");
         }
 
-        PKIFailureInfo  failInfo = tsResp.getFailInfo();
-        
+        PKIFailureInfo failInfo = tsResp.getFailInfo();
+
         if (failInfo == null)
         {
             fail("badPolicy - failInfo set to null.");
         }
-        
+
         if (failInfo.intValue() != PKIFailureInfo.unacceptedPolicy)
         {
             fail("badPolicy - wrong failure info returned.");
         }
     }
-    
+
     private void certReqTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -554,28 +661,28 @@ public class NewTSPTest
         TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(infoGeneratorBuilder.build(new JcaContentSignerBuilder("MD5withRSA").setProvider(BC).build(privateKey), cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
 
         tsTokenGen.addCertificates(certs);
-        
+
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        
+
         //
         // request with certReq false
         //
         reqGen.setCertReq(false);
-        
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
         TimeStampResponse tsResp = tsRespGen.generateGrantedResponse(request, new BigInteger("23"), new Date());
-        
+
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
-        
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
+
         assertNull(tsToken.getTimeStampInfo().getGenTimeAccuracy());  // check for abscence of accuracy
-        
+
         assertEquals("1.2", tsToken.getTimeStampInfo().getPolicy().getId());
-        
+
         try
         {
             tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
@@ -585,21 +692,21 @@ public class NewTSPTest
             fail("certReq(false) verification of token failed.");
         }
 
-        Store   respCerts = tsToken.getCertificates();
-        
-        Collection  certsColl = respCerts.getMatches(null);
-        
+        Store respCerts = tsToken.getCertificates();
+
+        Collection certsColl = respCerts.getMatches(null);
+
         if (!certsColl.isEmpty())
         {
             fail("certReq(false) found certificates in response.");
         }
     }
-    
-    
+
+
     private void tokenEncodingTest(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -608,10 +715,10 @@ public class NewTSPTest
 
         tsTokenGen.addCertificates(certs);
 
-        TimeStampRequestGenerator  reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest           request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+        TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
-        TimeStampResponse          tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date());
+        TimeStampResponse tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date());
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
@@ -619,16 +726,16 @@ public class NewTSPTest
 
         if (!Arrays.areEqual(tsResponse.getEncoded(), tsResp.getEncoded())
             || !Arrays.areEqual(tsResponse.getTimeStampToken().getEncoded(),
-                        tsResp.getTimeStampToken().getEncoded()))
+            tsResp.getTimeStampToken().getEncoded()))
         {
             fail();
         }
     }
-    
+
     private void testAccuracyZeroCerts(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -640,9 +747,9 @@ public class NewTSPTest
         tsTokenGen.setAccuracySeconds(1);
         tsTokenGen.setAccuracyMillis(2);
         tsTokenGen.setAccuracyMicros(3);
-        
+
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
@@ -650,10 +757,10 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider("BC").build(cert));
-        
+
         //
         // check validation
         //
@@ -663,34 +770,34 @@ public class NewTSPTest
         // check tstInfo
         //
         TimeStampTokenInfo tstInfo = tsToken.getTimeStampInfo();
-        
+
         //
         // check accuracy
         //
         GenTimeAccuracy accuracy = tstInfo.getGenTimeAccuracy();
-        
+
         assertEquals(1, accuracy.getSeconds());
         assertEquals(2, accuracy.getMillis());
         assertEquals(3, accuracy.getMicros());
-        
+
         assertEquals(new BigInteger("23"), tstInfo.getSerialNumber());
-        
+
         assertEquals("1.2", tstInfo.getPolicy().getId());
-        
+
         //
         // test certReq
         //
         Store store = tsToken.getCertificates();
-        
+
         Collection certificates = store.getMatches(null);
-        
+
         assertEquals(0, certificates.size());
     }
-    
+
     private void testAccuracyWithCertsAndOrdering(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -702,17 +809,17 @@ public class NewTSPTest
         tsTokenGen.setAccuracySeconds(3);
         tsTokenGen.setAccuracyMillis(1);
         tsTokenGen.setAccuracyMicros(2);
-        
+
         tsTokenGen.setOrdering(true);
-        
+
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        
+
         reqGen.setCertReq(true);
-        
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
 
         assertTrue(request.getCertReq());
-        
+
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
         TimeStampResponse tsResp;
@@ -728,10 +835,10 @@ public class NewTSPTest
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider("BC").build(cert));
-        
+
         //
         // check validation
         //
@@ -741,38 +848,38 @@ public class NewTSPTest
         // check tstInfo
         //
         TimeStampTokenInfo tstInfo = tsToken.getTimeStampInfo();
-        
+
         //
         // check accuracy
         //
         GenTimeAccuracy accuracy = tstInfo.getGenTimeAccuracy();
-        
+
         assertEquals(3, accuracy.getSeconds());
         assertEquals(1, accuracy.getMillis());
         assertEquals(2, accuracy.getMicros());
-        
+
         assertEquals(new BigInteger("23"), tstInfo.getSerialNumber());
-        
+
         assertEquals("1.2.3", tstInfo.getPolicy().getId());
-        
+
         assertEquals(true, tstInfo.isOrdered());
-        
+
         assertEquals(tstInfo.getNonce(), BigInteger.valueOf(100));
-        
+
         //
         // test certReq
         //
         Store store = tsToken.getCertificates();
-        
+
         Collection certificates = store.getMatches(null);
-        
+
         assertEquals(2, certificates.size());
-    }   
-    
+    }
+
     private void testNoNonse(
-        PrivateKey      privateKey,
+        PrivateKey privateKey,
         X509Certificate cert,
-        Store       certs)
+        Store certs)
         throws Exception
     {
         JcaSignerInfoGeneratorBuilder infoGeneratorBuilder = new JcaSignerInfoGeneratorBuilder(new JcaDigestCalculatorProviderBuilder().setProvider(BC).build());
@@ -780,22 +887,28 @@ public class NewTSPTest
         TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(infoGeneratorBuilder.build(new JcaContentSignerBuilder("MD5withRSA").setProvider(BC).build(privateKey), cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2.3"));
 
         tsTokenGen.addCertificates(certs);
-        
+
         TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
-        TimeStampRequest          request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20]);
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20]);
+
+        Set algorithms = new HashSet();
+
+        algorithms.add(TSPAlgorithms.SHA1);
+
+        request.validate(algorithms, new HashSet(), new HashSet());
 
         assertFalse(request.getCertReq());
-        
+
         TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
 
         TimeStampResponse tsResp = tsRespGen.generate(request, new BigInteger("24"), new Date());
 
         tsResp = new TimeStampResponse(tsResp.getEncoded());
 
-        TimeStampToken  tsToken = tsResp.getTimeStampToken();
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
 
         tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider("BC").build(cert));
-        
+
         //
         // check validation
         //
@@ -805,29 +918,112 @@ public class NewTSPTest
         // check tstInfo
         //
         TimeStampTokenInfo tstInfo = tsToken.getTimeStampInfo();
-        
+
         //
         // check accuracy
         //
         GenTimeAccuracy accuracy = tstInfo.getGenTimeAccuracy();
-        
+
         assertNull(accuracy);
-        
+
         assertEquals(new BigInteger("24"), tstInfo.getSerialNumber());
-        
+
         assertEquals("1.2.3", tstInfo.getPolicy().getId());
-        
+
         assertEquals(false, tstInfo.isOrdered());
-        
+
         assertNull(tstInfo.getNonce());
-        
+
         //
         // test certReq
         //
         Store store = tsToken.getCertificates();
-        
+
         Collection certificates = store.getMatches(null);
-        
+
         assertEquals(0, certificates.size());
-    } 
+    }
+
+    private void extensionTest(
+        PrivateKey privateKey,
+        X509Certificate cert,
+        Store certs)
+        throws Exception
+    {
+        TimeStampTokenGenerator tsTokenGen = new TimeStampTokenGenerator(
+            new JcaSimpleSignerInfoGeneratorBuilder().build("SHA1withRSA", privateKey, cert), new SHA1DigestCalculator(), new ASN1ObjectIdentifier("1.2"));
+
+        tsTokenGen.addCertificates(certs);
+
+        TimeStampRequestGenerator reqGen = new TimeStampRequestGenerator();
+
+        // test case only!!!
+        reqGen.setReqPolicy(Extension.noRevAvail);
+        // test case only!!!
+        reqGen.addExtension(Extension.biometricInfo, true, new DEROctetString(new byte[20]));
+
+        TimeStampRequest request = reqGen.generate(TSPAlgorithms.SHA1, new byte[20], BigInteger.valueOf(100));
+
+        try
+        {
+            request.validate(new HashSet(), new HashSet(), new HashSet());
+            fail("no exception");
+        }
+        catch (Exception e)
+        {
+            assertEquals(e.getMessage(), "request contains unknown algorithm");
+        }
+
+        Set algorithms = new HashSet();
+
+        algorithms.add(TSPAlgorithms.SHA1);
+
+        try
+        {
+            request.validate(algorithms, new HashSet(), new HashSet());
+            fail("no exception");
+        }
+        catch (Exception e)
+        {
+            assertEquals(e.getMessage(), "request contains unknown policy");
+        }
+
+        Set policies = new HashSet();
+
+        policies.add(Extension.noRevAvail);
+
+        try
+        {
+            request.validate(algorithms, policies, new HashSet());
+            fail("no exception");
+        }
+        catch (Exception e)
+        {
+            assertEquals(e.getMessage(), "request contains unknown extension");
+        }
+
+        Set extensions = new HashSet();
+
+        extensions.add(Extension.biometricInfo);
+
+        // should validate with full set
+        request.validate(algorithms, policies, extensions);
+
+        // should validate with null policy
+        request.validate(algorithms, null, extensions);
+
+        TimeStampResponseGenerator tsRespGen = new TimeStampResponseGenerator(tsTokenGen, TSPAlgorithms.ALLOWED);
+
+        TimeStampResponse tsResp = tsRespGen.generate(request, new BigInteger("23"), new Date());
+
+        tsResp = new TimeStampResponse(tsResp.getEncoded());
+
+        TimeStampToken tsToken = tsResp.getTimeStampToken();
+
+        tsToken.validate(new JcaSimpleSignerInfoVerifierBuilder().setProvider(BC).build(cert));
+
+        AttributeTable table = tsToken.getSignedAttributes();
+
+        assertNotNull("no signingCertificate attribute found", table.get(PKCSObjectIdentifiers.id_aa_signingCertificate));
+    }
 }
