@@ -25,6 +25,7 @@ import org.bouncycastle.asn1.cryptopro.CryptoProObjectIdentifiers;
 import org.bouncycastle.asn1.cryptopro.ECGOST3410NamedCurves;
 import org.bouncycastle.asn1.cryptopro.GOST3410PublicKeyAlgParameters;
 import org.bouncycastle.asn1.pkcs.PrivateKeyInfo;
+import org.bouncycastle.asn1.util.ASN1Dump;
 import org.bouncycastle.asn1.x509.AlgorithmIdentifier;
 import org.bouncycastle.asn1.x509.SubjectPublicKeyInfo;
 import org.bouncycastle.asn1.x9.X962Parameters;
@@ -41,6 +42,7 @@ import org.bouncycastle.jce.provider.BouncyCastleProvider;
 import org.bouncycastle.jce.spec.ECNamedCurveParameterSpec;
 import org.bouncycastle.jce.spec.ECNamedCurveSpec;
 import org.bouncycastle.math.ec.ECCurve;
+import org.bouncycastle.util.Strings;
 
 public class BCECGOST3410PrivateKey
     implements ECPrivateKey, org.bouncycastle.jce.interfaces.ECPrivateKey, PKCS12BagAttributeCarrier, ECPointEncoder
@@ -221,15 +223,22 @@ public class BCECGOST3410PrivateKey
 
             ASN1Encodable privKey = info.parsePrivateKey();
 
-            byte[] encVal = ASN1OctetString.getInstance(privKey).getOctets();
-            byte[] dVal = new byte[encVal.length];
-
-            for (int i = 0; i != encVal.length; i++)
+            if (privKey instanceof ASN1Integer)
             {
-                dVal[i] = encVal[encVal.length - 1 - i];
+                this.d = ASN1Integer.getInstance(privKey).getPositiveValue();
             }
+            else
+            {
+                byte[] encVal = ASN1OctetString.getInstance(privKey).getOctets();
+                byte[] dVal = new byte[encVal.length];
 
-            this.d = new BigInteger(1, dVal);
+                for (int i = 0; i != encVal.length; i++)
+                {
+                    dVal[i] = encVal[encVal.length - 1 - i];
+                }
+
+                this.d = new BigInteger(1, dVal);
+            }
         }
         else
         {
@@ -347,6 +356,7 @@ public class BCECGOST3410PrivateKey
         else
         {
             X962Parameters params;
+            int orderBitLength;
 
             if (ecSpec instanceof ECNamedCurveSpec)
             {
@@ -356,10 +366,12 @@ public class BCECGOST3410PrivateKey
                     curveOid = new ASN1ObjectIdentifier(((ECNamedCurveSpec)ecSpec).getName());
                 }
                 params = new X962Parameters(curveOid);
+                orderBitLength = ECUtil.getOrderBitLength(BouncyCastleProvider.CONFIGURATION, ecSpec.getOrder(), this.getS());
             }
             else if (ecSpec == null)
             {
                 params = new X962Parameters(DERNull.INSTANCE);
+                orderBitLength = ECUtil.getOrderBitLength(BouncyCastleProvider.CONFIGURATION, null, this.getS());
             }
             else
             {
@@ -373,6 +385,7 @@ public class BCECGOST3410PrivateKey
                     ecSpec.getCurve().getSeed());
 
                 params = new X962Parameters(ecP);
+                orderBitLength = ECUtil.getOrderBitLength(BouncyCastleProvider.CONFIGURATION, ecSpec.getOrder(), this.getS());
             }
 
             PrivateKeyInfo info;
@@ -380,11 +393,11 @@ public class BCECGOST3410PrivateKey
 
             if (publicKey != null)
             {
-                keyStructure = new org.bouncycastle.asn1.sec.ECPrivateKey(this.getS(), publicKey, params);
+                keyStructure = new org.bouncycastle.asn1.sec.ECPrivateKey(orderBitLength, this.getS(), publicKey, params);
             }
             else
             {
-                keyStructure = new org.bouncycastle.asn1.sec.ECPrivateKey(this.getS(), params);
+                keyStructure = new org.bouncycastle.asn1.sec.ECPrivateKey(orderBitLength, this.getS(), params);
             }
 
             try
@@ -494,7 +507,7 @@ public class BCECGOST3410PrivateKey
     public String toString()
     {
         StringBuffer buf = new StringBuffer();
-        String nl = System.getProperty("line.separator");
+        String nl = Strings.lineSeparator();
 
         buf.append("EC Private Key").append(nl);
         buf.append("             S: ").append(this.d.toString(16)).append(nl);
