@@ -29,10 +29,22 @@ public interface ConfigurableProvider
      */
     static final String DH_DEFAULT_PARAMS = "DhDefaultParams";
 
+    /**
+     * A set of OBJECT IDENTIFIERs representing acceptable named curves for imported keys.
+     */
+    static final String ACCEPTABLE_EC_CURVES = "acceptableEcCurves";
+
+    /**
+     * A set of OBJECT IDENTIFIERs to EC Curves providing local curve name mapping.
+     */
+    static final String ADDITIONAL_EC_PARAMETERS = "additionalEcParameters";
+
     void setParameter(String parameterName, Object parameter);
 
     void addAlgorithm(String key, String value);
 
+    void addAlgorithm(String type, ASN1ObjectIdentifier oid, String className);
+
     boolean hasAlgorithm(String type, String name);
 
     void addKeyInfoConverter(ASN1ObjectIdentifier oid, AsymmetricKeyInfoConverter keyInfoConverter);
