@@ -8,6 +8,7 @@ import java.security.MessageDigest;
 import java.security.NoSuchAlgorithmException;
 import java.security.NoSuchProviderException;
 import java.security.Provider;
+import java.security.SecureRandom;
 import java.security.Signature;
 import java.security.cert.CertificateException;
 import java.security.cert.CertificateFactory;
@@ -103,4 +104,10 @@ public class ProviderJcaJceHelper
     {
         return CertificateFactory.getInstance(algorithm, provider.getName());
     }
+
+    public SecureRandom createSecureRandom(String algorithm)
+        throws NoSuchAlgorithmException, NoSuchProviderException
+    {
+        return new SecureRandom();
+    }
 }
