@@ -1,15 +1,5 @@
 package org.bouncycastle.jce.provider.test;
 
-import org.bouncycastle.asn1.nist.NISTNamedCurves;
-import org.bouncycastle.asn1.sec.SECNamedCurves;
-import org.bouncycastle.asn1.teletrust.TeleTrusTNamedCurves;
-import org.bouncycastle.asn1.x9.X962NamedCurves;
-import org.bouncycastle.asn1.cryptopro.ECGOST3410NamedCurves;
-import org.bouncycastle.jce.provider.BouncyCastleProvider;
-import org.bouncycastle.jce.spec.ECNamedCurveSpec;
-import org.bouncycastle.util.test.SimpleTest;
-
-import javax.crypto.KeyAgreement;
 import java.math.BigInteger;
 import java.security.KeyFactory;
 import java.security.KeyPair;
@@ -22,12 +12,28 @@ import java.security.Signature;
 import java.security.interfaces.ECPrivateKey;
 import java.security.interfaces.ECPublicKey;
 import java.security.spec.ECGenParameterSpec;
+import java.security.spec.InvalidKeySpecException;
 import java.security.spec.PKCS8EncodedKeySpec;
 import java.security.spec.X509EncodedKeySpec;
+import java.util.Collections;
 import java.util.Enumeration;
+import java.util.HashSet;
 import java.util.Hashtable;
 import java.util.Set;
-import java.util.HashSet;
+
+import javax.crypto.KeyAgreement;
+
+import org.bouncycastle.asn1.ASN1ObjectIdentifier;
+import org.bouncycastle.asn1.cryptopro.ECGOST3410NamedCurves;
+import org.bouncycastle.asn1.nist.NISTNamedCurves;
+import org.bouncycastle.asn1.sec.SECNamedCurves;
+import org.bouncycastle.asn1.teletrust.TeleTrusTNamedCurves;
+import org.bouncycastle.asn1.x509.Extension;
+import org.bouncycastle.asn1.x9.X962NamedCurves;
+import org.bouncycastle.jcajce.provider.config.ConfigurableProvider;
+import org.bouncycastle.jce.provider.BouncyCastleProvider;
+import org.bouncycastle.jce.spec.ECNamedCurveSpec;
+import org.bouncycastle.util.test.SimpleTest;
 
 public class NamedCurveTest
     extends SimpleTest
@@ -282,6 +288,79 @@ public class NamedCurveTest
         }
     }
 
+    public void testAcceptable()
+        throws Exception
+    {
+        ECGenParameterSpec     ecSpec = new ECGenParameterSpec("P-256");
+        KeyPairGenerator       kpGen = KeyPairGenerator.getInstance("EC", "BC");
+
+        kpGen.initialize(ecSpec);
+
+        KeyPair kp = kpGen.generateKeyPair();
+
+        X509EncodedKeySpec pubSpec = new X509EncodedKeySpec(kp.getPublic().getEncoded());
+        PKCS8EncodedKeySpec privSpec = new PKCS8EncodedKeySpec(kp.getPrivate().getEncoded());
+
+        KeyFactory kf = KeyFactory.getInstance("EC", "BC");
+
+        ConfigurableProvider bcProv = ((ConfigurableProvider)Security.getProvider("BC"));
+
+        bcProv.setParameter(ConfigurableProvider.ACCEPTABLE_EC_CURVES, Collections.singleton(NISTNamedCurves.getOID("P-384")));
+
+        try
+        {
+            kf.generatePrivate(privSpec);
+            fail("no exception");
+        }
+        catch (InvalidKeySpecException e)
+        {
+            isTrue("wrong message", "encoded key spec not recognized: named curve not acceptable".equals(e.getMessage()));
+        }
+
+        try
+        {
+            kf.generatePublic(pubSpec);
+            fail("no exception");
+        }
+        catch (InvalidKeySpecException e)
+        {
+            isTrue("wrong message", "encoded key spec not recognized: named curve not acceptable".equals(e.getMessage()));
+        }
+
+        bcProv.setParameter(ConfigurableProvider.ACCEPTABLE_EC_CURVES, Collections.singleton(NISTNamedCurves.getOID("P-256")));
+
+        kf.generatePrivate(privSpec);
+        kf.generatePublic(pubSpec);
+
+        bcProv.setParameter(ConfigurableProvider.ACCEPTABLE_EC_CURVES, Collections.EMPTY_SET);
+
+        kf.generatePrivate(privSpec);
+        kf.generatePublic(pubSpec);
+    }
+
+    public void testAdditional()
+        throws Exception
+    {
+        ConfigurableProvider bcProv = ((ConfigurableProvider)Security.getProvider("BC"));
+        ASN1ObjectIdentifier bogusCurveID = Extension.auditIdentity;
+
+        bcProv.setParameter(ConfigurableProvider.ADDITIONAL_EC_PARAMETERS, Collections.singletonMap(bogusCurveID, NISTNamedCurves.getByName("P-384")));
+
+        KeyPairGenerator       kpGen = KeyPairGenerator.getInstance("EC", "BC");
+
+        kpGen.initialize(new ECGenParameterSpec(bogusCurveID.getId()));
+
+        KeyPair kp = kpGen.generateKeyPair();
+
+        X509EncodedKeySpec pubSpec = new X509EncodedKeySpec(kp.getPublic().getEncoded());
+        PKCS8EncodedKeySpec privSpec = new PKCS8EncodedKeySpec(kp.getPrivate().getEncoded());
+
+        KeyFactory kf = KeyFactory.getInstance("EC", "BC");
+
+        kf.generatePrivate(privSpec);
+        kf.generatePublic(pubSpec);
+    }
+
     public String getName()
     {
         return "NamedCurve";
@@ -329,6 +408,9 @@ public class NamedCurveTest
         {
             testECGOST((String)en.nextElement());
         }
+
+        testAcceptable();
+        testAdditional();
     }
     
     public static void main(
